import logging
from pathlib import Path, PosixPath
from collections import defaultdict
import tqdm

from ..dataset import GluonCVMotionDataset, get_vis_thumb_location
from ..utils.ingestion_utils import get_chunked_id_map


_log = logging.getLogger()
_log.setLevel(logging.DEBUG)


def get_data_gen(image_files=None):
    import imagededup.utils.data_generator as data_generator
    orig_datagen = data_generator.DataGenerator
    class PathNoName(PosixPath):
        @property
        def name(self):
            return str(self)

    class DataGenerator(orig_datagen):
        def _get_image_files(self) -> None:
            self.invalid_image_idx = []
            if image_files:
                self.image_files = [PathNoName(i) for i in image_files]
            else:
                self.image_files = sorted(
                    [
                        PathNoName(i.absolute())
                        for i in self.image_dir.rglob('*.jpg')
                        if not i.name.startswith('.')]
                )  # ignore hidden files
    return DataGenerator


def run_parallel(encoder, files):
    from imagededup.utils.general_utils import parallelise
    hashes = parallelise(encoder.encode_image, files, encoder.verbose)
    hash_dict = dict(zip(files, hashes))
    return hash_dict


def set_from_dict(d):
    vals = []
    for sid, v in d.items():
        if v:
            vals.append(sid)
        for x in v:
            dup_sid = x[0]
            vals.append(dup_sid)
    return set(vals)


def select_to_keep(samples, existing_dataset=None):
    if existing_dataset is None:
        existing_dataset = set()

    def get_dim(s): return s.width * s.height
    def get_duration(s): return len(s) / s.fps * 1000
    def cond_dim(sample, keep):
        return get_dim(sample) > get_dim(keep)
    def cond_duration(sample, keep):
        return get_dim(sample) >= (get_dim(keep) * 0.95) and \
               (get_duration(sample) - get_duration(keep) > 5000)

    keep = samples[0]
    for sample in samples[1:]:
        if sample.id in existing_dataset:
            keep = sample
        elif cond_dim(sample, keep) and not cond_duration(keep, sample):
            keep = sample
        elif cond_duration(sample, keep):
            keep = sample

    remove_dups = [s for s in samples if (s != keep) and (s.id not in existing_dataset)]
    return keep, remove_dups


def get_merged_dups(duplicates):
    def get_child_dups(path, dups_set=None):
        if dups_set is None:
            dups_set = set()
        dups_set.add(path)
        dups = duplicates[path]
        for dup in dups:
            if dup not in dups_set:
                get_child_dups(dup, dups_set)
        return dups_set

    new_duplicates = {}
    for path in duplicates:
        path_dups = get_child_dups(path) - {path}
        new_duplicates[path] = [(x, -1) for x in path_dups]
    return new_duplicates


def get_unique_dups(duplicates, image_dir, dataset, thumb_to_sample_id, sample_id_to_thumb):
    selected_dups = set()
    all_dups = set()
    dups_map = defaultdict(list)
    dups_keep_lookup = {}

    for path, dups in duplicates.items():
        rel_path = str(Path(path).relative_to(image_dir))
        if rel_path not in selected_dups:
            for dup in dups:
                dup_path, score = dup
                rel_dup_path = str(Path(dup_path).relative_to(image_dir))
                all_dups.add(rel_dup_path)
                dups_map[rel_path].append((rel_dup_path, score))
            thumb_paths = [rel_path] + [str(Path(dup[0]).relative_to(image_dir)) for dup in dups
                                        if dup not in selected_dups]
            samples = [dataset[thumb_to_sample_id[p]] for p in thumb_paths]
            keep, remove_dups = select_to_keep(samples)
            remove_dups_thumbs = [sample_id_to_thumb[s.id] for s in remove_dups]
            selected_dups.update(remove_dups_thumbs)
            for dup in remove_dups_thumbs:
                dups_keep_lookup[dup] = sample_id_to_thumb[keep.id]
    return all_dups, dups_map, selected_dups


def get_files_to_inspect(dups_map, dataset, thumb_to_sample_id, orig_dataset=None, id_map=None):
    inspect_map = defaultdict(list)
    def get_samp(sid):
        sid = thumb_to_sample_id[sid]
        if orig_dataset and id_map and sid in id_map:
            return orig_dataset[id_map[sid]]
        else:
            return dataset[sid]

    for sid, v in dups_map.items():
        samp = get_samp(sid)
        for (dup_sid, score) in v:
            dup_samp = get_samp(dup_sid)
            len_diff = abs(len(samp) - len(dup_samp))
            if len_diff > 10:
                inspect_map[sid].append((dup_sid, score, len_diff))
    return inspect_map


def get_encodings(im_paths, method="cnn"):
    import imagededup.utils.data_generator as data_generator
    from imagededup.methods import PHash, CNN
    if method == "cnn":
        data_generator.DataGenerator = get_data_gen(im_paths)
        deduper = CNN()
        encodings = deduper.encode_images('/')
    elif method == "phash":
        deduper = PHash()
        encodings = run_parallel(deduper, im_paths)
    else:
        raise ValueError(f"Method {method} unknown")
    return encodings, deduper


def get_duplicates(im_paths, method="cnn", encodings=None, deduper=None, **kwargs):
    from imagededup.methods import PHash, CNN
    if encodings is None:
        encodings, deduper = get_encodings(im_paths, method)
    if method == "cnn":
        deduper = deduper or CNN()
        assert isinstance(deduper, CNN)
        thresh = "min_similarity_threshold"
        if thresh not in kwargs:
            kwargs[thresh] = 0.9
    elif method == "phash":
        deduper = deduper or PHash()
        assert isinstance(deduper, PHash)
        thresh = "max_distance_threshold"
        if thresh not in kwargs:
            kwargs[thresh] = 10
    else:
        raise ValueError(f"Method {method} unknown")
    duplicates = deduper.find_duplicates(encoding_map=encodings, scores=True, **kwargs)
    return duplicates, encodings, deduper


def main(anno_path, out_anno=None, orig_anno_file=None):
    dataset = GluonCVMotionDataset(anno_path)
    if out_anno is None:
        anno_path = Path(anno_path)
        out_anno = anno_path.with_name("{}_dedupe{}".format(anno_path.stem, anno_path.suffix))
    out_dataset = GluonCVMotionDataset(out_anno, dataset.root_path, load_anno=False)
    out_dataset.metadata = dataset.metadata
    out_dataset.description += "; deduped"
    samples = dataset.samples

    if orig_anno_file:
        orig_dataset = GluonCVMotionDataset(orig_anno_file, root_path=dataset.root_path)
        orig_id_map = get_chunked_id_map(dataset, orig_dataset, chunk_subdirs=["vid_chunks_2m_1p"],
                                         exclude_unchunked=True)
    else:
        orig_dataset = None
        orig_id_map = {}

    image_dir = Path(dataset.cache_root_path) / "thumbnails"

    thumb_paths = []
    thumb_to_sample_id = {}
    sample_id_to_thumb = {}
    for sample_id, sample in tqdm.tqdm(samples, mininterval=1):
        thumb_path = get_vis_thumb_location(sample)
        if not Path(thumb_path).exists():
            _log.warning("Missing thumbnail: {}".format(thumb_path))
            continue
        thumb_paths.append(thumb_path)
        thumb_rel_path = str(Path(thumb_path).relative_to(image_dir))
        thumb_to_sample_id[thumb_rel_path] = sample_id
        sample_id_to_thumb[sample_id] = thumb_rel_path

    def im_filter(path):
        if path.suffix.lower() not in [".jpeg", ".jpg", ".png"]:
            return False
        return True


    duplicates, _, _ = get_duplicates(thumb_paths, method="cnn", min_similarity_threshold=0.83)

    dup_set = set_from_dict(duplicates)
    addn_args = [image_dir, dataset, thumb_to_sample_id, sample_id_to_thumb]
    all_dups, dups_map, selected_dups = get_unique_dups(duplicates, *addn_args)

    print(len(selected_dups))
    dup_sample_ids = [thumb_to_sample_id[thumb_path] \
                     for thumb_path in selected_dups \
                     if "videos_MEVA" not in thumb_path \
                        and "Offline_CCTV_Footage" not in thumb_path]
    print(len(dup_sample_ids))
    for sample_id, sample in tqdm.tqdm(samples):
        if sample_id not in dup_sample_ids:
                out_dataset.add_sample(sample)

    out_dataset.dump()


if __name__ == "__main__":
    import fire
    fire.Fire(main)



"""Code adapted from https://github.com/open-mmlab/mmaction
and https://github.com/bryanyzhu/two-stream-pytorch"""

import random
from PIL import Image, ImageOps
import numpy as np
import numbers
import math
import cv2
import collections
import PIL


class Transform:

    def __call__(self, data):
        raise NotImplemented


class Compose(object):
    """Composes several transforms together.
    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.
    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, data):
        for t in self.transforms:
            data = t(data)
        return data


###################################################################################
#   Array Related Transformations
###################################################################################


class ToNumpy(Transform):
    def __call__(self, image):
        arr = np.array(image, dtype=np.float32)
        return arr


class StackImage(Transform):
    """
    This transformation stacks all images in nested lists into one list
    It works in a depth-first manner
    """

    def __init__(self, level=2):
        self.level = level

    def __call__(self, images):
        out_array = []

        def traverse(array, depth, max_depth):
            if depth != max_depth:
                for x in array:
                    traverse(x, depth + 1, max_depth)
            else:
                out_array.extend(array)

        traverse(images, 1, self.level)
        return out_array


class GroupApply(Transform):
    def __init__(self, transform):
        self.trans = transform

    def __call__(self, input_list):
        return [self.trans(x) for x in input_list]


class ToCNNInput(Transform):
    def __call__(self, array):

        if len(array.shape) == 3:
            return array.transpose((2, 0, 1))
        else:
            return array


class Normalize(Transform):
    def __init__(self, mean=128, std=128):
        self.mean = mean
        self.std = std

    def __call__(self, array):
        return (array - self.mean) / self.std


################################################################################################
# Basic Transformations
################################################################################################


class IdentityTransform(Transform):

    def __call__(self, data):
        return data


class RandomCrop(Transform):

    def __init__(self, size):

        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def __call__(self, img):

        w, h = (img[0].shape[0], img[0].shape[1])
        th, tw = self.size
        clip_len = 64

        if tw > w or th > h:
            raise ValueError("RandomCrop Failed: crop size is larger than image size")

        x1 = random.randint(0, w - tw)
        y1 = random.randint(0, h - th)

        buffer = np.array(img)

        buffer = buffer[:,
                 x1:x1 + th,
                 y1:y1 + tw, :]

        return buffer


class CenterCrop(Transform):

    def __init__(self, size):

        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size

    def __call__(self, img):

        w, h = img.size
        th, tw = self.size

        if tw > w or th > h:
            raise ValueError("CenterCrop Failed: crop size is larger than image size")

        x1 = int(round((w - tw) / 2.))
        y1 = int(round((h - th) / 2.))

        return img.crop((x1, y1, x1 + tw, y1 + th))


class Scale(Transform):
    """Rescale the input PIL.Image to the given size.
    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (w, h), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``
    """

    def __init__(self, size, interpolation=Image.BILINEAR):

        assert isinstance(size, int) or (isinstance(size, collections.Iterable) and len(size) == 2)
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img):

        if isinstance(self.size, int):
            w, h = img.size
            if (w <= h and w == self.size) or (h <= w and h == self.size):
                return img
            if w < h:
                ow = self.size
                oh = int(self.size * h / w)
                return img.resize((ow, oh), self.interpolation)
            else:
                oh = self.size
                ow = int(self.size * w / h)
                return img.resize((ow, oh), self.interpolation)
        else:
            return img.resize(self.size, self.interpolation)


################################################################################################
# Group Transformations: apply transformation to sequence of images
################################################################################################


class GroupRandomCrop(Transform):

    def __init__(self, size):
        self.worker = RandomCrop(size)

    def __call__(self, img_group):
        return [self.worker(img) for img in img_group]


class GroupCenterCrop(Transform):

    def __init__(self, size):
        self.worker = CenterCrop(size)

    def __call__(self, img_group):
        return [self.worker(img) for img in img_group]


class GroupRandomHorizontalFlip(Transform):
    """Randomly horizontally flips the given PIL.Image with a given probability
    """

    def __init__(self, is_flow=False, prob=0.5):
        self.is_flow = is_flow
        self.prob = prob

    def __call__(self, img_group, is_flow=False):
        v = random.random()
        if v < self.prob:

            ret = [img.transpose(Image.FLIP_LEFT_RIGHT) for img in img_group]
            if self.is_flow:
                for i in range(0, len(ret), 2):
                    ret[i] = ImageOps.invert(ret[i])  # invert flow pixel values when flipping
            return ret
        else:
            return img_group


class AlternateHorizontalFlip(Transform):
    """
    Input is a list of len the number of segments. Each element is a list of PIL images of length num. of segment frames (snippet length)
    All frames are flipped horizontally for every other segment
    """

    def __init__(self, is_flow=False):
        self.is_flow = is_flow

    def __call__(self, frames):

        def horizontal_flip_img_list(img_group, is_flow):
            ret = [img.transpose(Image.FLIP_LEFT_RIGHT) for img in img_group]
            if is_flow:
                for i in range(0, len(ret), 2):
                    ret[i] = ImageOps.invert(ret[i])  # invert flow pixel values when flipping
            return ret

        n_segments = len(frames)
        for s_idx in range(n_segments):
            if s_idx % 2 == 1:
                frames[s_idx] = horizontal_flip_img_list(frames[s_idx], self.is_flow)
        return frames


class GroupNormalize(Transform):

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor):
        rep_mean = self.mean * (tensor.shape[0] // len(self.mean))
        rep_std = self.std * (tensor.shape[0] // len(self.std))

        for t, m, s in zip(tensor, rep_mean, rep_std):
            t.__sub__(m).__div__(s)

        return tensor


class GroupScale(Transform):

    def __init__(self, size, interpolation=Image.BILINEAR):
        self.worker = Scale(size, interpolation)

    def __call__(self, img_group):
        return [self.worker(img) for img in img_group]


class GroupOverSample(Transform):

    def __init__(self, crop_size, scale_size=None):
        self.crop_size = crop_size if not isinstance(crop_size, int) else (crop_size, crop_size)

        if scale_size is not None:
            self.scale_worker = GroupScale(scale_size)
        else:
            self.scale_worker = None

    def __call__(self, img_group):

        if self.scale_worker is not None:
            img_group = self.scale_worker(img_group)

        image_w, image_h = img_group[0].size
        crop_w, crop_h = self.crop_size

        offsets = GroupMultiScaleCrop.fill_fix_offset(False, image_w, image_h, crop_w, crop_h)
        oversample_group = list()
        for o_w, o_h in offsets:
            normal_group = list()
            flip_group = list()
            for i, img in enumerate(img_group):
                crop = img.crop((o_w, o_h, o_w + crop_w, o_h + crop_h))
                normal_group.append(crop)
                flip_crop = crop.copy().transpose(Image.FLIP_LEFT_RIGHT)

                if img.mode == 'L' and i % 2 == 0:
                    flip_group.append(ImageOps.invert(flip_crop))
                else:
                    flip_group.append(flip_crop)
            oversample_group.extend(normal_group)
            oversample_group.extend(flip_group)
        return oversample_group


class GroupMultiScaleCrop(Transform):

    def __init__(self, input_size, scales=None, max_distort=1, fix_crop=True, more_fix_crop=True):
        self.scales = scales if scales is not None else [1, 875, .75, .66]
        self.max_distort = max_distort
        self.fix_crop = fix_crop
        self.more_fix_crop = more_fix_crop
        self.input_size = input_size if not isinstance(input_size, int) else [input_size, input_size]
        self.interpolation = Image.BILINEAR

    def __call__(self, img_group):
        im_size = img_group[0].size

        crop_w, crop_h, offset_w, offset_h = self._sample_crop_size(im_size)
        crop_img_group = [img.crop((offset_w, offset_h, offset_w + crop_w, offset_h + crop_h)) for img in img_group]
        ret_img_group = [img.resize((self.input_size[0], self.input_size[1]), self.interpolation)
                         for img in crop_img_group]
        return ret_img_group

    def _sample_crop_size(self, im_size):
        image_w, image_h = im_size[0], im_size[1]

        # find a crop size
        base_size = min(image_w, image_h)
        crop_sizes = [int(base_size * x) for x in self.scales]
        crop_h = [self.input_size[1] if abs(x - self.input_size[1]) < 3 else x for x in crop_sizes]
        crop_w = [self.input_size[0] if abs(x - self.input_size[0]) < 3 else x for x in crop_sizes]

        pairs = []
        for i, h in enumerate(crop_h):
            for j, w in enumerate(crop_w):
                if abs(i - j) <= self.max_distort:
                    pairs.append((w, h))

        crop_pair = random.choice(pairs)
        if not self.fix_crop:
            w_offset = random.randint(0, image_w - crop_pair[0])
            h_offset = random.randint(0, image_h - crop_pair[1])
        else:
            w_offset, h_offset = self._sample_fix_offset(image_w, image_h, crop_pair[0], crop_pair[1])

        return crop_pair[0], crop_pair[1], w_offset, h_offset

    def _sample_fix_offset(self, image_w, image_h, crop_w, crop_h):
        offsets = self.fill_fix_offset(self.more_fix_crop, image_w, image_h, crop_w, crop_h)
        return random.choice(offsets)

    @staticmethod
    def fill_fix_offset(more_fix_crop, image_w, image_h, crop_w, crop_h):
        w_step = (image_w - crop_w) // 4
        h_step = (image_h - crop_h) // 4

        ret = list()
        ret.append((0, 0))  # upper left
        ret.append((4 * w_step, 0))  # upper right
        ret.append((0, 4 * h_step))  # lower left
        ret.append((4 * w_step, 4 * h_step))  # lower right
        ret.append((2 * w_step, 2 * h_step))  # center

        if more_fix_crop:
            ret.append((0, 2 * h_step))  # center left
            ret.append((4 * w_step, 2 * h_step))  # center right
            ret.append((2 * w_step, 4 * h_step))  # lower center
            ret.append((2 * w_step, 0 * h_step))  # upper center

            ret.append((1 * w_step, 1 * h_step))  # upper left quarter
            ret.append((3 * w_step, 1 * h_step))  # upper right quarter
            ret.append((1 * w_step, 3 * h_step))  # lower left quarter
            ret.append((3 * w_step, 3 * h_step))  # lower righ quarter

        return ret


class GroupRandomSizedCrop(Transform):
    """Random crop the given PIL.Image to a random size of (0.08 to 1.0) of the original size
    and and a random aspect ratio of 3/4 to 4/3 of the original aspect ratio
    This is popularly used to train the Inception networks
    size: size of the smaller edge
    interpolation: Default: PIL.Image.BILINEAR
    """

    def __init__(self, size, interpolation=Image.BILINEAR):
        self.size = size
        self.interpolation = interpolation

    def __call__(self, img_group):
        for attempt in range(10):
            area = img_group[0].size[0] * img_group[0].size[1]
            target_area = random.uniform(0.08, 1.0) * area
            aspect_ratio = random.uniform(3. / 4, 4. / 3)

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if random.random() < 0.5:
                w, h = h, w

            if w <= img_group[0].size[0] and h <= img_group[0].size[1]:
                x1 = random.randint(0, img_group[0].size[0] - w)
                y1 = random.randint(0, img_group[0].size[1] - h)
                found = True
                break
        else:
            found = False
            x1 = 0
            y1 = 0

        if found:
            out_group = list()
            for img in img_group:
                img = img.crop((x1, y1, x1 + w, y1 + h))
                assert (img.size == (w, h))
                out_group.append(img.resize((self.size, self.size), self.interpolation))
            return out_group
        else:
            # Fallback
            scale = GroupScale(self.size, interpolation=self.interpolation)
            crop = GroupRandomCrop(self.size)
            return crop(scale(img_group))


class Stack(Transform):

    def __init__(self, roll=False):
        self.roll = roll

    def __call__(self, img_group):
        if img_group[0].mode == 'L':
            return np.concatenate([np.expand_dims(x, 2) for x in img_group], axis=2)
        elif img_group[0].mode == 'RGB':
            if self.roll:
                return np.concatenate([np.array(x)[:, :, ::-1] for x in img_group], axis=2)
            else:
                return np.concatenate(img_group, axis=2)


def parse_mean_std(mean, std, is_flow=False):
    if mean is None:
        mean = np.array([0.485, 0.456, 0.406], dtype=np.float32) * 255
        if is_flow:
            mean = np.float32(128)

    if std is None:
        std = np.array([0.229, 0.224, 0.225], dtype=np.float32) * 255
        if is_flow:
            std = std.mean()
    return mean, std


class RandomHorizontalFlip(object):
    """Horizontally flip the list of given images randomly
    with a probability 0.5
    """

    def __call__(self, clip):
        """
        Args:
        img (PIL.Image or numpy.ndarray): List of images to be cropped
        in format (h, w, c) in numpy.ndarray

        Returns:
        PIL.Image or numpy.ndarray: Randomly flipped clip
        """
        if random.random() < 0.5:
            if isinstance(clip[0], np.ndarray):
                return [np.fliplr(img) for img in clip]
            elif isinstance(clip[0], PIL.Image.Image):
                return [
                    img.transpose(PIL.Image.FLIP_LEFT_RIGHT) for img in clip
                ]
            else:
                raise TypeError('Expected numpy.ndarray or PIL.Image' +
                                ' but got list of {0}'.format(type(clip[0])))
        return clip


class Resize(object):
    """Resizes a list of (H x W x C) numpy.ndarray to the final size

    The larger the original image is, the more times it takes to
    interpolate

    Args:
    interpolation (str): Can be one of 'nearest', 'bilinear'
    defaults to nearest
    size (tuple): (widht, height)
    """

    def __init__(self, size, interpolation='nearest'):
        self.size = size
        self.interpolation = interpolation

    def __call__(self, clip):
        resized = self.resize_clip(
            clip, self.size, interpolation=self.interpolation)
        return resized

    def resize_clip(self, clip, size, interpolation='bilinear'):
        if isinstance(clip[0], np.ndarray):
            if isinstance(size, numbers.Number):
                im_h, im_w, im_c = clip[0].shape
                # Min spatial dim already matches minimal size
                if (im_w <= im_h and im_w == size) or (im_h <= im_w and im_h == size):
                    return clip
                new_h, new_w = self.get_resize_sizes(im_h, im_w, size)
                size = (new_w, new_h)
            else:
                size = size[1], size[0]
            if interpolation == 'bilinear':
                np_inter = cv2.INTER_LINEAR
            else:
                np_inter = cv2.INTER_NEAREST
            scaled = [cv2.resize(img, size, interpolation=np_inter) for img in clip]
        elif isinstance(clip[0], PIL.Image.Image):
            if isinstance(size, numbers.Number):
                im_w, im_h = clip[0].size
                # Min spatial dim already matches minimal size
                if (im_w <= im_h and im_w == size) or (im_h <= im_w and im_h == size):
                    return clip
                new_h, new_w = self.get_resize_sizes(im_h, im_w, size)
                size = (new_w, new_h)
            else:
                size = size[1], size[0]
            if interpolation == 'bilinear':
                pil_inter = PIL.Image.NEAREST
            else:
                pil_inter = PIL.Image.BILINEAR
            scaled = [img.resize(size, pil_inter) for img in clip]
        else:
            raise TypeError('Expected numpy.ndarray or PIL.Image' +
                            'but got list of {0}'.format(type(clip[0])))
        return scaled

    def get_resize_sizes(self, im_h, im_w, size):
        if im_w < im_h:
            ow = size
            oh = int(size * im_h / im_w)
        else:
            oh = size
            ow = int(size * im_w / im_h)
        return oh, ow


def get_standard_tsn_training_transform(input_size, scales=(1, .875, .75, .66), is_flow=False, mean=None, std=None):
    """
    This function return a composed transform ready to be pluged into the TSN dataset for training
    :param std:
    :param mean:
    :param input_size: the expected networks input size
    :param scales: scales in scale jittering
    :param is_flow: whether to flip the image in the flow style
    :return:
    """

    mean, std = parse_mean_std(mean, std, is_flow)

    if is_flow:
        scales = (1, .875, .75)

    image_wise = Compose([
        ToNumpy(),
        Normalize(mean=mean, std=std),
        ToCNNInput()
    ])
    transform = Compose(
        [
            StackImage(2 if not is_flow else 3),
            Compose([GroupMultiScaleCrop(input_size, scales),
                     GroupRandomHorizontalFlip(is_flow=is_flow)]),
            GroupApply(image_wise),
        ]
    )

    return transform


def get_standard_tsn_validation_transform(input_size, is_flow=False, mean=None, std=None):
    """
    This function returns a composed transform ready to be pluged into the TSN dataset for validation
    :param std:
    :param mean:
    :param input_size: the expected networks input size
    :param is_flow: whether to flip the image in the flow style
    :return:
    """

    mean, std = parse_mean_std(mean, std, is_flow)

    if input_size != (224, 224):
        resize = (int(340 * input_size[0] / 224), int(256 * input_size[1] / 224))
    else:
        resize = None

    image_wise = Compose([
        ToNumpy(),
        Normalize(mean=mean, std=std),
        ToCNNInput()
    ])
    transform = Compose(
        [
            StackImage(2 if not is_flow else 3),
            GroupScale(resize) if resize is not None else IdentityTransform(),
            Compose([GroupCenterCrop(input_size)]),
            GroupApply(image_wise),
        ]
    )

    return transform


def get_standard_tsn_testing_transform(oversample, input_size, is_flow=False, mean=None, std=None):
    """
    This function returns a composed transform ready to be pluged into the TSN dataset for testing
    """

    mean, std = parse_mean_std(mean, std, is_flow)

    if input_size != (224, 224):
        resize = (int(340 * input_size[0] / 224), int(256 * input_size[1] / 224))
    else:
        resize = None

    image_wise = Compose([
        ToNumpy(),
        Normalize(mean=mean, std=std),
        ToCNNInput()
    ])

    if not oversample:
        transform = Compose(
            [
                StackImage(2 if not is_flow else 3),
                GroupScale(resize) if resize is not None else IdentityTransform(),
                Compose([GroupCenterCrop(input_size)]),
                GroupApply(image_wise),
            ]
        )
    else:
        transform = Compose(
            [
                StackImage(2 if not is_flow else 3),
                GroupScale(resize) if resize is not None else IdentityTransform(),
                Compose([GroupOverSample(input_size)]),
                GroupApply(image_wise),
            ]
        )
    return transform


def get_3D_testing_transform(alternate_flip=True, center_crop=True, input_size=(224, 224), is_flow=False, mean=None,
                             std=None):
    mean, std = parse_mean_std(mean, std, is_flow)

    image_wise = Compose([
        ToNumpy(),
        Normalize(mean=mean, std=std),
        ToCNNInput()
    ])

    transform = Compose(
        [
            AlternateHorizontalFlip(is_flow) if alternate_flip else IdentityTransform(),
            StackImage(2 if not is_flow else 3),
            GroupCenterCrop(input_size) if center_crop else IdentityTransform(),
            GroupApply(image_wise),
        ]
    )

    return transform


def resize_clip_bb(clip, size, anno, interpolation='bilinear'):
    '''
    Resize clip and associate bounding boxes
    '''
    if isinstance(clip[0], np.ndarray):
        if isinstance(size, numbers.Number):
            im_h, im_w, im_c = clip[0].shape
            # Min spatial dim already matches minimal size
            if (im_w <= im_h and im_w == size) or (im_h <= im_w
                                                   and im_h == size):
                return clip, anno
            new_h, new_w, scale = get_resize_sizes_scale(im_h, im_w, size)
            size = (new_w, new_h)
        else:
            size = size[1], size[0]
        if interpolation == 'bilinear':
            np_inter = cv2.INTER_LINEAR
        else:
            np_inter = cv2.INTER_NEAREST
        scaled = [
            cv2.resize(img, size, interpolation=np_inter) for img in clip
        ]
    elif isinstance(clip[0], PIL.Image.Image):
        if isinstance(size, numbers.Number):
            im_w, im_h = clip[0].size
            # Min spatial dim already matches minimal size
            if (im_w <= im_h and im_w == size) or (im_h <= im_w
                                                   and im_h == size):
                return clip, anno
            new_h, new_w, scale = get_resize_sizes_scale(im_h, im_w, size)
            size = (new_w, new_h)
        else:
            size = size[1], size[0]
        if interpolation == 'bilinear':
            pil_inter = PIL.Image.NEAREST
        else:
            pil_inter = PIL.Image.BILINEAR
        scaled = [img.resize(size, pil_inter) for img in clip]
    else:
        raise TypeError('Expected numpy.ndarray or PIL.Image' +
                        'but got list of {0}'.format(type(clip[0])))

    for i in range(anno.__len__()):
        anno[i] = anno[i] * scale

    return scaled, anno


def get_resize_sizes_scale(im_h, im_w, size):
    if im_w < im_h:
        ow = size
        oh = int(size * im_h / im_w)
        scale = size / im_w
    else:
        oh = size
        ow = int(size * im_w / im_h)
        scale = size / im_h
    return oh, ow, scale


def randomcrop_clip_bb(clip, anno, crop_size):
    '''
        crop clip and associate bounding boxes
    '''
    h, w = (crop_size, crop_size)
    if isinstance(clip[0], np.ndarray):
        im_h, im_w, im_c = clip[0].shape
    elif isinstance(clip[0], PIL.Image.Image):
        im_w, im_h = clip[0].size
    else:
        raise TypeError('Expected numpy.ndarray or PIL.Image' +
                        'but got list of {0}'.format(type(clip[0])))
    if w > im_w or h > im_h:
        error_msg = (
            'Initial image size should be larger then '
            'cropped size but got cropped sizes : ({w}, {h}) while '
            'initial image is ({im_w}, {im_h})'.format(
                im_w=im_w, im_h=im_h, w=w, h=h))
        raise ValueError(error_msg)

    x1 = random.randint(0, im_w - w)
    y1 = random.randint(0, im_h - h)
    cropped = crop_clip_bb(clip, y1, x1, h, w)
    c_anno = anno
    for i in range(anno.__len__()):
        c_anno[i][0] = np.max((int(anno[i][0] - x1), 0))
        c_anno[i][1] = np.max((int(anno[i][1] - y1), 0))
        c_anno[i][2] = np.max((int(anno[i][2]), anno[i][0] + anno[i][2] - x1 - c_anno[i][0]))
        c_anno[i][3] = np.max((int(anno[i][3]), anno[i][1] + anno[i][3] - y1 - c_anno[i][1]))
    return cropped, c_anno


def crop_clip_bb(clip, min_h, min_w, h, w):
    if isinstance(clip[0], np.ndarray):
        cropped = [img[min_h:min_h + h, min_w:min_w + w, :] for img in clip]

    elif isinstance(clip[0], PIL.Image.Image):
        cropped = [
            img.crop((min_w, min_h, min_w + w, min_h + h)) for img in clip
        ]
    else:
        raise TypeError('Expected numpy.ndarray or PIL.Image' +
                        'but got list of {0}'.format(type(clip[0])))
    return cropped


def random_horizontal_flip_bb(clip, anno):
    '''
        crop clip and associate bounding boxes
    '''
    import random
    prob = random.uniform(0, 1)
    if prob < 0.5:
        return clip, anno

    if isinstance(clip[0], np.ndarray):
        im_h, im_w, im_c = clip[0].shape
    elif isinstance(clip[0], PIL.Image.Image):
        im_w, im_h = clip[0].size
    else:
        raise TypeError('Expected numpy.ndarray or PIL.Image' +
                        'but got list of {0}'.format(type(clip[0])))
    mid = im_w // 2
    for i in range(len(clip)):
        clip[i] = clip[i][:, ::-1, :]
    for i in range(anno.__len__()):
        anno[i][0] = np.max((int(mid - (anno[i][0] + anno[i][2] - mid)), 0))
        anno[i][2] = np.min((im_w - anno[i][0], anno[i][2]))
    return clip, anno


def normalize_pytorch(buffer):
    # Normalize the buffer
    buffer = buffer / 255.0
    buffer[:, :, :, 0] = (buffer[:, :, :, 0] - 0.45) / 0.225
    buffer[:, :, :, 1] = (buffer[:, :, :, 1] - 0.45) / 0.225
    buffer[:, :, :, 2] = (buffer[:, :, :, 2] - 0.45) / 0.225

    return buffer


def to_tensor_pytorch(buffer):
    # convert from [D, H, W, C] format to [C, D, H, W] (what PyTorch uses)
    # D = Depth (in this case, time), H = Height, W = Width, C = Channels
    return buffer.transpose((3, 0, 1, 2))



# coding: utf-8
# pylint: disable=wrong-import-position
"""GluonCV: a deep learning vision toolkit powered by Gluon."""
from __future__ import absolute_import

from .check import _deprecate_python2
from .check import _require_mxnet_version, _require_pytorch_version

__version__ = '0.10.0'

_deprecate_python2()

# optionally depend on mxnet or pytorch
try:
    _require_mxnet_version('1.4.0', '2.0.0')
    from . import data
    from . import model_zoo
    from . import nn
    from . import utils
    from . import loss
except ImportError:
    try:
        _require_pytorch_version('1.4.0', '2.0.0')
        from .torch import data
        from .torch import model_zoo
        from .torch import utils
    except ImportError:
        raise ImportError('Unable to import modules due to missing `mxnet` & `torch`. '
                          'You should install at least one deep learning framework.')



"""Utility functions for version checking."""
import warnings

__all__ = ['check_version']

def check_version(min_version, warning_only=False):
    """Check the version of gluoncv satisfies the provided minimum version.
    An exception is thrown if the check does not pass.

    Parameters
    ----------
    min_version : str
        Minimum version
    warning_only : bool
        Printing a warning instead of throwing an exception.
    """
    from .. import __version__
    from distutils.version import LooseVersion
    bad_version = LooseVersion(__version__) < LooseVersion(min_version)
    if bad_version:
        msg = 'Installed GluonCV version (%s) does not satisfy the ' \
              'minimum required version (%s)'%(__version__, min_version)
        if warning_only:
            warnings.warn(msg)
        else:
            raise AssertionError(msg)



# pylint: disable=missing-function-docstring, unused-argument
"""
Non-local module from Non-local Neural Networks
CVPR 2018, https://arxiv.org/abs/1711.07971
Code adapted from https://github.com/open-mmlab/mmaction
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import BatchNorm3d


def build_nonlocal_block(cfg):
    """ Build nonlocal block from
    `"Non-local Neural Networks"
    <https://arxiv.org/abs/1711.07971>`_ paper.
    Code adapted from mmaction.
    """
    assert isinstance(cfg, dict)
    cfg_ = cfg.copy()
    return NonLocal(**cfg_)


class NonLocal(nn.Module):
    """
    Non-local module from Non-local Neural Networks
    CVPR 2018, https://arxiv.org/abs/1711.07971
    """
    def __init__(self, in_channels=1024, nonlocal_type="gaussian", dim=3,
                 embed=True, embed_dim=None, sub_sample=False, use_bn=True,
                 norm_layer=BatchNorm3d, norm_kwargs=None, **kwargs):
        super(NonLocal, self).__init__()

        assert nonlocal_type in ['gaussian', 'dot', 'concat']
        self.nonlocal_type = nonlocal_type
        self.embed = embed
        self.embed_dim = embed_dim if embed_dim is not None else in_channels // 2
        self.sub_sample = sub_sample
        self.use_bn = use_bn

        if self.embed:
            if dim == 2:
                self.theta = nn.Conv2d(in_channels=in_channels,
                                       out_channels=self.embed_dim,
                                       kernel_size=(1, 1),
                                       stride=(1, 1),
                                       padding=(0, 0))
                self.phi = nn.Conv2d(in_channels=in_channels,
                                     out_channels=self.embed_dim,
                                     kernel_size=(1, 1),
                                     stride=(1, 1),
                                     padding=(0, 0))
                self.g = nn.Conv2d(in_channels=in_channels,
                                   out_channels=self.embed_dim,
                                   kernel_size=(1, 1),
                                   stride=(1, 1),
                                   padding=(0, 0))
            elif dim == 3:
                self.theta = nn.Conv3d(in_channels=in_channels,
                                       out_channels=self.embed_dim,
                                       kernel_size=(1, 1, 1),
                                       stride=(1, 1, 1),
                                       padding=(0, 0, 0))
                self.phi = nn.Conv3d(in_channels=in_channels,
                                     out_channels=self.embed_dim,
                                     kernel_size=(1, 1, 1),
                                     stride=(1, 1, 1),
                                     padding=(0, 0, 0))
                self.g = nn.Conv3d(in_channels=in_channels,
                                   out_channels=self.embed_dim,
                                   kernel_size=(1, 1, 1),
                                   stride=(1, 1, 1),
                                   padding=(0, 0, 0))

        if self.nonlocal_type == 'concat':
            if dim == 2:
                self.concat_proj = nn.Sequential(
                    nn.Conv2d(in_channels=self.embed_dim * 2,
                              out_channels=1,
                              kernel_size=(1, 1),
                              stride=(1, 1),
                              padding=(0, 0)),
                    nn.ReLU(inplace=True))
            elif dim == 3:
                self.concat_proj = nn.Sequential(
                    nn.Conv3d(in_channels=self.embed_dim * 2,
                              out_channels=1,
                              kernel_size=(1, 1, 1),
                              stride=(1, 1, 1),
                              padding=(0, 0, 0)),
                    nn.ReLU(inplace=True))

        if sub_sample:
            if dim == 2:
                self.max_pool = nn.MaxPool2d(pool_size=(2, 2))
            elif dim == 3:
                self.max_pool = nn.MaxPool3d(pool_size=(1, 2, 2))
            self.sub_phi = nn.Sequential(self.phi, self.max_pool)
            self.sub_g = nn.Sequential(self.g, self.max_pool)

        if dim == 2:
            self.W = nn.Conv2d(in_channels=self.embed_dim,
                               out_channels=in_channels,
                               kernel_size=(1, 1),
                               stride=(1, 1),
                               padding=(0, 0))
        elif dim == 3:
            self.W = nn.Conv3d(in_channels=self.embed_dim,
                               out_channels=in_channels,
                               kernel_size=(1, 1, 1),
                               stride=(1, 1, 1),
                               padding=(0, 0, 0))

        if use_bn:
            # TODO: need to add zero initialized BN, also the conv output
            self.bn = norm_layer(num_features=in_channels,
                                 momentum=0.9,
                                 **({} if norm_kwargs is None else norm_kwargs))
            self.W_bn = nn.Sequential(self.W, self.bn)

    def forward(self, x):
        if self.embed:
            theta = self.theta(x)
            if self.sub_sample:
                phi = self.sub_phi(x)
                g = self.sub_g(x)
            else:
                phi = self.phi(x)
                g = self.g(x)
        else:
            theta = x
            phi = x
            g = x

        theta_shape_5d = theta.shape
        if self.nonlocal_type == 'gaussian':
            # reshape [BxCxTxHxW] to [BxCxTHW]
            theta = theta.view(theta_shape_5d[0], theta_shape_5d[1], -1)
            phi = phi.view(theta_shape_5d[0], theta_shape_5d[1], -1)
            g = g.view(theta_shape_5d[0], theta_shape_5d[1], -1)

            theta_phi = torch.bmm(theta.transpose(1, 2), phi)
            theta_phi = theta_phi * (self.embed_dim ** -.5)
            attn = F.softmax(theta_phi, dim=-1)
        elif self.non_local_type == 'concat':
            raise NotImplementedError
        elif self.non_local_type == 'dot':
            raise NotImplementedError
        else:
            raise NotImplementedError

        y = torch.bmm(g, attn.transpose(1, 2))
        y = y.view(theta_shape_5d)

        if self.use_bn:
            z = self.W_bn(y) + x
        else:
            z = self.W(y) + x
        return z



# pylint: disable=arguments-differ, unused-argument
"""Samplers for positive/negative/ignore sample selections.
This module is used to select samples during training.
Based on different strategies, we would like to choose different number of
samples as positive, negative or ignore(don't care). The purpose is to alleviate
unbalanced training target in some circumstances.
The output of sampler is an NDArray of the same shape as the matching results.
Note: 1 for positive, -1 for negative, 0 for ignore.
"""
from __future__ import absolute_import

import numpy as np
import mxnet as mx
from mxnet import gluon, nd


class NaiveSampler(gluon.HybridBlock):
    """A naive sampler that take all existing matching results.
    There is no ignored sample in this case.
    """

    def __init__(self):
        super(NaiveSampler, self).__init__()

    def hybrid_forward(self, F, x):
        """Hybrid forward"""
        marker = F.ones_like(x)
        y = F.where(x >= 0, marker, marker * -1)
        return y


class OHEMSampler(gluon.Block):
    """A sampler implementing Online Hard-negative mining.
    As described in paper https://arxiv.org/abs/1604.03540.

    Parameters
    ----------
    ratio : float
        Ratio of negative vs. positive samples. Values >= 1.0 is recommended.
    min_samples : int, default 0
        Minimum samples to be selected regardless of positive samples.
        For example, if positive samples is 0, we sometimes still want some num_negative
        samples to be selected.
    thresh : float, default 0.5
        IOU overlap threshold of selected negative samples. IOU must not exceed
        this threshold such that good matching anchors won't be selected as
        negative samples.

    """

    def __init__(self, ratio, min_samples=0, thresh=0.5):
        super(OHEMSampler, self).__init__()
        assert ratio > 0, "OHEMSampler ratio must > 0, {} given".format(ratio)
        self._ratio = ratio
        self._min_samples = min_samples
        self._thresh = thresh

    # pylint: disable=arguments-differ
    def forward(self, x, logits, ious):
        """Forward"""
        F = nd
        num_positive = F.sum(x > -1, axis=1)
        num_negative = self._ratio * num_positive
        num_total = x.shape[1]  # scalar
        num_negative = F.minimum(F.maximum(self._min_samples, num_negative),
                                 num_total - num_positive)
        positive = logits.slice_axis(axis=2, begin=1, end=None)
        background = logits.slice_axis(axis=2, begin=0, end=1).reshape((0, -1))
        maxval = positive.max(axis=2)
        esum = F.exp(logits - maxval.reshape((0, 0, 1))).sum(axis=2)
        score = -F.log(F.exp(background - maxval) / esum)
        mask = F.ones_like(score) * -1
        score = F.where(x < 0, score, mask)  # mask out positive samples
        if len(ious.shape) == 3:
            ious = F.max(ious, axis=2)
        score = F.where(ious < self._thresh, score, mask)  # mask out if iou is large
        argmaxs = F.argsort(score, axis=1, is_ascend=False)

        # neg number is different in each batch, using dynamic numpy operations.
        y = np.zeros(x.shape)
        y[np.where(x.asnumpy() >= 0)] = 1  # assign positive samples
        argmaxs = argmaxs.asnumpy()
        for i, num_neg in zip(range(x.shape[0]), num_negative.asnumpy().astype(np.int32)):
            indices = argmaxs[i, :num_neg]
            y[i, indices.astype(np.int32)] = -1  # assign negative samples
        return F.array(y, ctx=x.context)


class QuotaSampler(gluon.Block):
    """Sampler that handles limited quota for positive and negative samples.

    Parameters
    ----------
    num_sample : int, default is 128
        Number of samples for RCNN targets.
    pos_iou_thresh : float, default is 0.5
        Proposal whose IOU larger than ``pos_iou_thresh`` is regarded as positive samples.
    neg_iou_thresh_high : float, default is 0.5
        Proposal whose IOU smaller than ``neg_iou_thresh_high``
        and larger than ``neg_iou_thresh_low``
        is regarded as negative samples.
        Proposals with IOU in between ``pos_iou_thresh`` and ``neg_iou_thresh`` are
        ignored.
    neg_iou_thresh_low : float, default is 0.0
        See ``neg_iou_thresh_high``.
    pos_ratio : float, default is 0.25
        ``pos_ratio`` defines how many positive samples (``pos_ratio * num_sample``) is
        to be sampled.
    neg_ratio : float or None
        ``neg_ratio`` defines how many negative samples (``pos_ratio * num_sample``) is
        to be sampled. If ``None`` is provided, it equals to ``1 - pos_ratio``.
    fill_negative : bool
        If ``True``, negative samples will fill the gap caused by insufficient positive samples.
        For example, if ``num_sample`` is 100, ``pos_ratio`` and ``neg_ratio`` are both ``0.5``.
        Available positive sample and negative samples are 10 and 10000, which are typical values.
        Now, the output positive samples is 10(intact), since it's smaller than ``50(100 * 0.5)``,
        the negative samples will fill the rest ``40`` slots.
        If ``fill_negative == False``, the ``40`` slots is filled with ``-1(ignore)``.

    """

    def __init__(self, num_sample, pos_thresh, neg_thresh_high, neg_thresh_low=-np.inf,
                 pos_ratio=0.5, neg_ratio=None, fill_negative=True):
        super(QuotaSampler, self).__init__()
        self._fill_negative = fill_negative
        self._num_sample = num_sample
        if neg_ratio is None:
            self._neg_ratio = 1. - pos_ratio
        self._pos_ratio = pos_ratio
        assert (self._neg_ratio + self._pos_ratio) <= 1.0, (
            "Positive and negative ratio {} exceed 1".format(self._neg_ratio + self._pos_ratio))
        self._pos_thresh = min(1., max(0., pos_thresh))
        self._neg_thresh_high = min(1., max(0., neg_thresh_high))
        self._neg_thresh_low = neg_thresh_low

    def forward(self, matches, ious):
        """Quota Sampler

        Parameters:
        ----------
        matches : NDArray or Symbol
            Matching results, positive number for positive matching, -1 for not matched.
        ious : NDArray or Symbol
            IOU overlaps with shape (N, M), batching is supported.

        Returns:
        --------
        NDArray or Symbol
            Sampling results with same shape as ``matches``.
            1 for positive, -1 for negative, 0 for ignore.

        """
        F = mx.nd
        max_pos = int(round(self._pos_ratio * self._num_sample))
        max_neg = int(self._neg_ratio * self._num_sample)
        results = []
        for i in range(matches.shape[0]):
            # init with 0s, which are ignored
            result = F.zeros_like(matches[0])
            # positive samples
            ious_max = ious.max(axis=-1)[i]
            result = F.where(matches[i] >= 0, F.ones_like(result), result)
            result = F.where(ious_max >= self._pos_thresh, F.ones_like(result), result)
            # negative samples with label -1
            neg_mask = ious_max < self._neg_thresh_high
            neg_mask = neg_mask * (ious_max >= self._neg_thresh_low)
            result = F.where(neg_mask, F.ones_like(result) * -1, result)

            # re-balance if number of positive or negative exceed limits
            result = result.asnumpy()
            num_pos = int((result > 0).sum())
            if num_pos > max_pos:
                disable_indices = np.random.choice(
                    np.where(result > 0)[0], size=(num_pos - max_pos), replace=False)
                result[disable_indices] = 0  # use 0 to ignore
            num_neg = int((result < 0).sum())
            if self._fill_negative:
                # if pos_sample is less than quota, we can have negative samples filling the gap
                max_neg = max(self._num_sample - min(num_pos, max_pos), max_neg)
            if num_neg > max_neg:
                disable_indices = np.random.choice(
                    np.where(result < 0)[0], size=(num_neg - max_neg), replace=False)
                result[disable_indices] = 0
            results.append(mx.nd.array(result))

        return mx.nd.stack(*results, axis=0)


class QuotaSamplerOp(mx.operator.CustomOp):
    """Sampler that handles limited quota for positive and negative samples.

    This is a custom Operator used inside HybridBlock.

    Parameters
    ----------
    num_sample : int, default is 128
        Number of samples for RCNN targets.
    pos_iou_thresh : float, default is 0.5
        Proposal whose IOU larger than ``pos_iou_thresh`` is regarded as positive samples.
    neg_iou_thresh_high : float, default is 0.5
        Proposal whose IOU smaller than ``neg_iou_thresh_high``
        and larger than ``neg_iou_thresh_low``
        is regarded as negative samples.
        Proposals with IOU in between ``pos_iou_thresh`` and ``neg_iou_thresh`` are
        ignored.
    neg_iou_thresh_low : float, default is 0.0
        See ``neg_iou_thresh_high``.
    pos_ratio : float, default is 0.25
        ``pos_ratio`` defines how many positive samples (``pos_ratio * num_sample``) is
        to be sampled.
    neg_ratio : float or None
        ``neg_ratio`` defines how many negative samples (``pos_ratio * num_sample``) is
        to be sampled. If ``None`` is provided, it equals to ``1 - pos_ratio``.
    fill_negative : bool
        If ``True``, negative samples will fill the gap caused by insufficient positive samples.
        For example, if ``num_sample`` is 100, ``pos_ratio`` and ``neg_ratio`` are both ``0.5``.
        Available positive sample and negative samples are 10 and 10000, which are typical values.
        Now, the output positive samples is 10(intact), since it's smaller than ``50(100 * 0.5)``,
        the negative samples will fill the rest ``40`` slots.
        If ``fill_negative == False``, the ``40`` slots is filled with ``-1(ignore)``.

    """

    def __init__(self, num_sample, pos_thresh, neg_thresh_high=0.5, neg_thresh_low=-np.inf,
                 pos_ratio=0.5, neg_ratio=None, fill_negative=True):
        super(QuotaSamplerOp, self).__init__()
        self._num_sample = num_sample
        self._fill_negative = fill_negative
        if neg_ratio is None:
            self._neg_ratio = 1. - pos_ratio
        self._pos_ratio = pos_ratio
        assert (self._neg_ratio + self._pos_ratio) <= 1.0, (
            "Positive and negative ratio {} exceed 1".format(self._neg_ratio + self._pos_ratio))
        self._pos_thresh = min(1., max(0., pos_thresh))
        self._neg_thresh_high = min(1., max(0., neg_thresh_high))
        self._neg_thresh_low = neg_thresh_low

    def forward(self, is_train, req, in_data, out_data, aux):
        """Quota Sampler

        Parameters:
        ----------
        in_data: array-like of Symbol
            [matches, ious], see below.
        matches : NDArray or Symbol
            Matching results, positive number for positive matching, -1 for not matched.
        ious : NDArray or Symbol
            IOU overlaps with shape (N, M), batching is supported.

        Returns:
        --------
        NDArray or Symbol
            Sampling results with same shape as ``matches``.
            1 for positive, -1 for negative, 0 for ignore.

        """
        matches = in_data[0]
        ious = in_data[1]
        F = mx.nd
        max_pos = int(round(self._pos_ratio * self._num_sample))
        max_neg = int(self._neg_ratio * self._num_sample)
        for i in range(matches.shape[0]):
            # init with 0s, which are ignored
            result = F.zeros_like(matches[i])
            # negative samples with label -1
            ious_max = ious.max(axis=-1)[i]
            neg_mask = ious_max < self._neg_thresh_high
            neg_mask = neg_mask * (ious_max >= self._neg_thresh_low)
            result = F.where(neg_mask, F.ones_like(result) * -1, result)
            # positive samples
            result = F.where(matches[i] >= 0, F.ones_like(result), result)
            result = F.where(ious_max >= self._pos_thresh, F.ones_like(result), result)

            # re-balance if number of positive or negative exceed limits
            result = result.asnumpy()
            num_pos = int((result > 0).sum())
            if num_pos > max_pos:
                disable_indices = np.random.choice(
                    np.where(result > 0)[0], size=(num_pos - max_pos), replace=False)
                result[disable_indices] = 0  # use 0 to ignore
            num_neg = int((result < 0).sum())
            if self._fill_negative:
                # if pos_sample is less than quota, we can have negative samples filling the gap
                max_neg = max(self._num_sample - min(num_pos, max_pos), max_neg)
            if num_neg > max_neg:
                disable_indices = np.random.choice(
                    np.where(result < 0)[0], size=(num_neg - max_neg), replace=False)
                result[disable_indices] = 0  # use 0 to ignore

            self.assign(out_data[0][i], req[0], mx.nd.array(result))

    def backward(self, req, out_grad, in_data, out_data, in_grad, aux):
        self.assign(in_grad[0], req[0], 0)
        self.assign(in_grad[1], req[1], 0)


@mx.operator.register('quota_sampler')
class QuotaSamplerProp(mx.operator.CustomOpProp):
    """Property for QuotaSampleOp.

    Parameters
    ----------
    num_sample : int, default is 128
        Number of samples for RCNN targets.
    pos_iou_thresh : float, default is 0.5
        Proposal whose IOU larger than ``pos_iou_thresh`` is regarded as positive samples.
    neg_iou_thresh_high : float, default is 0.5
        Proposal whose IOU smaller than ``neg_iou_thresh_high``
        and larger than ``neg_iou_thresh_low``
        is regarded as negative samples.
        Proposals with IOU in between ``pos_iou_thresh`` and ``neg_iou_thresh`` are
        ignored.
    neg_iou_thresh_low : float, default is 0.0
        See ``neg_iou_thresh_high``.
    pos_ratio : float, default is 0.25
        ``pos_ratio`` defines how many positive samples (``pos_ratio * num_sample``) is
        to be sampled.
    neg_ratio : float or None
        ``neg_ratio`` defines how many negative samples (``pos_ratio * num_sample``) is
        to be sampled. If ``None`` is provided, it equals to ``1 - pos_ratio``.
    fill_negative : bool
        If ``True``, negative samples will fill the gap caused by insufficient positive samples.
        For example, if ``num_sample`` is 100, ``pos_ratio`` and ``neg_ratio`` are both ``0.5``.
        Available positive sample and negative samples are 10 and 10000, which are typical values.
        Now, the output positive samples is 10(intact), since it's smaller than ``50(100 * 0.5)``,
        the negative samples will fill the rest ``40`` slots.
        If ``fill_negative == False``, the ``40`` slots is filled with ``-1(ignore)``.

    """

    def __init__(self, num_sample, pos_thresh, neg_thresh_high=0.5, neg_thresh_low=0.,
                 pos_ratio=0.5, neg_ratio=None, fill_negative=True):
        super(QuotaSamplerProp, self).__init__(need_top_grad=False)
        self.num_sample = int(num_sample)
        self.pos_thresh = float(pos_thresh)
        self.neg_thresh_high = float(neg_thresh_high)
        self.neg_thresh_low = float(neg_thresh_low)
        self.pos_ratio = float(pos_ratio)
        self.neg_ratio = None if neg_ratio is None else float(neg_ratio)
        self.fill_negative = bool(fill_negative)

    def list_arguments(self):
        return ['matches', 'ious']

    def list_outputs(self):
        return ['output']

    def infer_shape(self, in_shape):
        return in_shape, [in_shape[0]], []

    def infer_type(self, in_type):
        return [in_type[0], in_type[0]], [in_type[0]], []

    # pylint: disable=unused-argument
    def create_operator(self, ctx, in_shapes, in_dtypes):
        return QuotaSamplerOp(self.num_sample, self.pos_thresh, self.neg_thresh_high,
                              self.neg_thresh_low, self.pos_ratio, self.neg_ratio,
                              self.fill_negative)



# pylint: disable=missing-module-docstring
from .resnet_encoder import ResnetEncoder
from .depth_decoder import DepthDecoder
from .pose_decoder import PoseDecoder
from .monodepth2 import *
from .monodepth2_posenet import *



# pylint: disable=unused-argument,abstract-method,missing-docstring,arguments-differ
"""Image Cascade Network (ICNet)
ICNet for Real-Time Semantic Segmentation on High-Resolution Images, ECCV 2018
https://hszhao.github.io/projects/icnet/
Code partially borrowed from https://github.com/lxtGH/Fast_Seg/blob/master/libs/models/ICNet.py.
"""
from __future__ import division
from mxnet.gluon import nn
from mxnet.context import cpu
from mxnet.gluon.nn import HybridBlock
from gluoncv.model_zoo.segbase import SegBaseModel
from gluoncv.model_zoo.pspnet import _PSPHead

__all__ = ['ICNet', 'get_icnet', 'get_icnet_resnet50_citys', 'get_icnet_resnet50_mhpv1']

class ICNet(SegBaseModel):
    r"""Image Cascade Network (ICNet)

    Parameters
    ----------
    nclass : int
        Number of categories for the training dataset.
    backbone : string
        Pre-trained dilated backbone network type (default:'resnet50'; 'resnet50',
        'resnet101' or 'resnet152').
    norm_layer : object
        Normalization layer used in backbone network (default: :class:`mxnet.gluon.nn.BatchNorm`;
        for Synchronized Cross-GPU BachNormalization).
    norm_kwargs : dict
        Additional `norm_layer` arguments, for example `num_devices=4`
        for :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.
    pretrained_base : bool or str
        Refers to if the backbone is pretrained or not. If `True`,
        model weights of a model that was trained on ImageNet is loaded.

    Reference:

        Hengshuang Zhao, Xiaojuan Qi, Xiaoyong Shen, Jianping Shi and Jiaya Jia.
        "ICNet for Real-Time Semantic Segmentation on High-Resolution Images." *ECCV*, 2018

    Examples
    --------
    >>> model = ICNet(nclass=19, backbone='resnet50')
    >>> print(model)
    """

    def __init__(self, nclass, backbone='resnet50', aux=False, ctx=cpu(), pretrained_base=True,
                 height=None, width=None, base_size=520, crop_size=480, lr_mult=10, **kwargs):
        super(ICNet, self).__init__(nclass, aux=aux, backbone=backbone, ctx=ctx,
                                    base_size=base_size, crop_size=crop_size,
                                    pretrained_base=pretrained_base, **kwargs)

        height = height if height is not None else crop_size
        width = width if width is not None else crop_size
        self._up_kwargs = {'height': height, 'width': width}
        self.base_size = base_size
        self.crop_size = crop_size

        with self.name_scope():
            # large resolution branch
            self.conv_sub1 = nn.HybridSequential()
            with self.conv_sub1.name_scope():
                self.conv_sub1.add(ConvBnRelu(3, 32, 3, 2, 1, **kwargs),
                                   ConvBnRelu(32, 32, 3, 2, 1, **kwargs),
                                   ConvBnRelu(32, 64, 3, 2, 1, **kwargs))
            self.conv_sub1.initialize(ctx=ctx)
            self.conv_sub1.collect_params().setattr('lr_mult', lr_mult)

            # small and medium resolution branches, backbone comes from segbase.py
            self.psp_head = _PSPHead(nclass,
                                     feature_map_height=self._up_kwargs['height'] // 32,
                                     feature_map_width=self._up_kwargs['width'] // 32,
                                     **kwargs)
            self.psp_head.block = self.psp_head.block[:-1]
            self.psp_head.initialize(ctx=ctx)
            self.psp_head.collect_params().setattr('lr_mult', lr_mult)

            # ICNet head
            self.head = _ICHead(nclass=nclass,
                                height=self._up_kwargs['height'],
                                width=self._up_kwargs['width'],
                                **kwargs)
            self.head.initialize(ctx=ctx)
            self.head.collect_params().setattr('lr_mult', lr_mult)

            # reduce conv
            self.conv_sub4 = ConvBnRelu(512, 256, 1, **kwargs)
            self.conv_sub4.initialize(ctx=ctx)
            self.conv_sub4.collect_params().setattr('lr_mult', lr_mult)

            self.conv_sub2 = ConvBnRelu(512, 256, 1, **kwargs)
            self.conv_sub2.initialize(ctx=ctx)
            self.conv_sub2.collect_params().setattr('lr_mult', lr_mult)

    def hybrid_forward(self, F, x):
        # large resolution branch
        x_sub1_out = self.conv_sub1(x)

        # medium resolution branch
        x_sub2 = F.contrib.BilinearResize2D(x,
                                            height=self._up_kwargs['height'] // 2,
                                            width=self._up_kwargs['width'] // 2)
        x = self.conv1(x_sub2)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x_sub2_out = self.layer2(x)

        # small resolution branch
        x_sub4 = F.contrib.BilinearResize2D(x_sub2_out,
                                            height=self._up_kwargs['height'] // 32,
                                            width=self._up_kwargs['width'] // 32)
        x = self.layer3(x_sub4)
        x = self.layer4(x)
        x_sub4_out = self.psp_head(x)

        # reduce conv
        x_sub4_out = self.conv_sub4(x_sub4_out)
        x_sub2_out = self.conv_sub2(x_sub2_out)

        # ICNet head
        res = self.head(x_sub1_out, x_sub2_out, x_sub4_out)
        return res

    def demo(self, x):
        return self.predict(x)

    def predict(self, x):
        h, w = x.shape[2:]
        self._up_kwargs['height'] = h
        self._up_kwargs['width'] = w

        import mxnet.ndarray as F
        x_sub1_out = self.conv_sub1(x)

        x_sub2 = F.contrib.BilinearResize2D(x,
                                            height=self._up_kwargs['height'] // 2,
                                            width=self._up_kwargs['width'] // 2)
        x = self.conv1(x_sub2)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x_sub2_out = self.layer2(x)

        x_sub4 = F.contrib.BilinearResize2D(x_sub2_out,
                                            height=self._up_kwargs['height'] // 32,
                                            width=self._up_kwargs['width'] // 32)
        x = self.layer3(x_sub4)
        x = self.layer4(x)
        x_sub4_out = self.psp_head.demo(x)

        x_sub4_out = self.conv_sub4(x_sub4_out)
        x_sub2_out = self.conv_sub2(x_sub2_out)
        res = self.head.demo(x_sub1_out, x_sub2_out, x_sub4_out)
        return res[0]

class _ICHead(HybridBlock):
    # pylint: disable=redefined-outer-name
    def __init__(self, nclass, height=None, width=None,
                 norm_layer=nn.BatchNorm, **kwargs):
        super(_ICHead, self).__init__()

        self._up_kwargs = {'height': height, 'width': width}
        self.cff_12 = CascadeFeatureFusion(low_channels=128,
                                           high_channels=64,
                                           out_channels=128,
                                           nclass=nclass,
                                           height=height // 8,
                                           width=width // 8,
                                           norm_layer=norm_layer,
                                           **kwargs)
        self.cff_24 = CascadeFeatureFusion(low_channels=256,
                                           high_channels=256,
                                           out_channels=128,
                                           nclass=nclass,
                                           height=height // 16,
                                           width=width // 16,
                                           norm_layer=norm_layer,
                                           **kwargs)

        with self.name_scope():
            self.conv_cls = nn.Conv2D(in_channels=128, channels=nclass,
                                      kernel_size=1, use_bias=False)

    def hybrid_forward(self, F, x_sub1, x_sub2, x_sub4):
        outputs = []

        x_cff_24, x_24_cls = self.cff_24(x_sub4, x_sub2)
        outputs.append(x_24_cls)
        x_cff_12, x_12_cls = self.cff_12(x_cff_24, x_sub1)
        outputs.append(x_12_cls)

        up_x2 = F.contrib.BilinearResize2D(x_cff_12,
                                           height=self._up_kwargs['height'] // 4,
                                           width=self._up_kwargs['width'] // 4)
        up_x2 = self.conv_cls(up_x2)
        outputs.append(up_x2)

        up_x8 = F.contrib.BilinearResize2D(up_x2,
                                           height=self._up_kwargs['height'],
                                           width=self._up_kwargs['width'])
        outputs.append(up_x8)

        # 1 -> 1/4 -> 1/8 -> 1/16
        outputs.reverse()
        return tuple(outputs)

    def demo(self, x_sub1, x_sub2, x_sub4):
        outputs = []

        x_cff_24, x_24_cls = self.cff_24.demo(x_sub4, x_sub2)
        outputs.append(x_24_cls)
        x_cff_12, x_12_cls = self.cff_12.demo(x_cff_24, x_sub1)
        outputs.append(x_12_cls)

        import mxnet.ndarray as F
        up_x2 = F.contrib.BilinearResize2D(x_cff_12,
                                           height=x_cff_12.shape[2] * 2,
                                           width=x_cff_12.shape[3] * 2)
        up_x2 = self.conv_cls(up_x2)
        outputs.append(up_x2)

        up_x8 = F.contrib.BilinearResize2D(up_x2,
                                           height=up_x2.shape[2] * 4,
                                           width=up_x2.shape[3] * 4)
        outputs.append(up_x8)

        # 1 -> 1/4 -> 1/8 -> 1/16
        outputs.reverse()
        return tuple(outputs)

class CascadeFeatureFusion(HybridBlock):
    def __init__(self, low_channels, high_channels, out_channels,
                 nclass, height=None, width=None,
                 norm_layer=nn.BatchNorm, **kwargs):
        super(CascadeFeatureFusion, self).__init__()
        self._up_kwargs = {'height': height, 'width': width}

        with self.name_scope():
            self.conv_low = nn.HybridSequential()
            with self.conv_low.name_scope():
                self.conv_low.add(nn.Conv2D(in_channels=low_channels, channels=out_channels,
                                            kernel_size=3, padding=2, dilation=2, use_bias=False))
                self.conv_low.add(norm_layer(in_channels=out_channels))

            self.conv_hign = nn.HybridSequential()
            with self.conv_hign.name_scope():
                self.conv_hign.add(nn.Conv2D(in_channels=high_channels, channels=out_channels,
                                             kernel_size=1, use_bias=False))
                self.conv_hign.add(norm_layer(in_channels=out_channels))

            self.conv_low_cls = nn.Conv2D(in_channels=out_channels, channels=nclass,
                                          kernel_size=1, use_bias=False)

    def hybrid_forward(self, F, x_low, x_high):
        x_low = F.contrib.BilinearResize2D(x_low,
                                           height=self._up_kwargs['height'],
                                           width=self._up_kwargs['width'])
        x_low = self.conv_low(x_low)
        x_high = self.conv_hign(x_high)

        x = x_low + x_high
        x = F.relu(x)

        x_low_cls = self.conv_low_cls(x_low)
        return x, x_low_cls

    def demo(self, x_low, x_high):
        self._up_kwargs['height'] = x_high.shape[2]
        self._up_kwargs['width'] = x_high.shape[3]

        import mxnet.ndarray as F
        x_low = F.contrib.BilinearResize2D(x_low,
                                           height=self._up_kwargs['height'],
                                           width=self._up_kwargs['width'])
        x_low = self.conv_low(x_low)
        x_high = self.conv_hign(x_high)

        x = x_low + x_high
        x = F.relu(x)

        x_low_cls = self.conv_low_cls(x_low)
        return x, x_low_cls

class ConvBnRelu(HybridBlock):
    def __init__(self, in_planes, out_planes, ksize, stride=1, pad=0, dilation=1,
                 groups=1, has_bn=True, norm_layer=nn.BatchNorm, bn_eps=1e-5,
                 has_relu=True, has_bias=False, **kwargs):
        super(ConvBnRelu, self).__init__()
        with self.name_scope():
            self.conv = nn.Conv2D(in_channels=in_planes, channels=out_planes,
                                  kernel_size=ksize, padding=pad, strides=stride,
                                  dilation=dilation, groups=groups, use_bias=has_bias)
            self.has_bn = has_bn
            self.has_relu = has_relu

            if self.has_bn:
                self.bn = norm_layer(in_channels=out_planes, epsilon=bn_eps)
            if self.has_relu:
                self.relu = nn.Activation('relu')

    def hybrid_forward(self, F, x):
        x = self.conv(x)
        if self.has_bn:
            x = self.bn(x)
        if self.has_relu:
            x = self.relu(x)
        return x

    def demo(self, x):
        x = self.conv(x)
        if self.has_bn:
            x = self.bn(x)
        if self.has_relu:
            x = self.relu(x)
        return x

def get_icnet(dataset='citys', backbone='resnet50', pretrained=False,
              root='~/.mxnet/models', pretrained_base=True, ctx=cpu(0), **kwargs):
    r"""Image Cascade Network

    Parameters
    ----------
    dataset : str, default citys
        The dataset that model pretrained on. (default: cityscapes)
    backbone : string
        Pre-trained dilated backbone network type (default:'resnet50'; 'resnet50',
        'resnet101' or 'resnet152').
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default '~/.mxnet/models'
        Location for keeping the model parameters.
    pretrained_base : bool or str, default True
        This will load pretrained backbone network, that was trained on ImageNet.

    """
    acronyms = {
        'pascal_voc': 'voc',
        'pascal_aug': 'voc',
        'ade20k': 'ade',
        'coco': 'coco',
        'citys': 'citys',
        'mhpv1': 'mhpv1',
    }
    from ..data import datasets
    # infer number of classes
    model = ICNet(datasets[dataset].NUM_CLASS, backbone=backbone,
                  pretrained_base=pretrained_base, ctx=ctx, **kwargs)
    model.classes = datasets[dataset].CLASSES

    if pretrained:
        from .model_store import get_model_file
        model.load_parameters(get_model_file('icnet_%s_%s' % (backbone, acronyms[dataset]),
                                             tag=pretrained, root=root), ctx=ctx)
    return model


def get_icnet_resnet50_citys(**kwargs):
    r"""Image Cascade Network

    Parameters
    ----------
    dataset : str, default citys
        The dataset that model pretrained on. (default: cityscapes)
    backbone : string
        Pre-trained dilated backbone network type (default:'resnet50').

    """
    return get_icnet(dataset='citys', backbone='resnet50', **kwargs)


def get_icnet_resnet50_mhpv1(**kwargs):
    r"""Image Cascade Network

    Parameters
    ----------
    dataset : str, default citys
        The dataset that model pretrained on. (default: cityscapes)
    backbone : string
        Pre-trained dilated backbone network type (default:'resnet50').

    """
    return get_icnet(dataset='mhpv1', backbone='resnet50', **kwargs)



# pylint: disable=wildcard-import
"""Video action recognition, HMDB51 dataset.
http://serre-lab.clps.brown.edu/resource/hmdb-a-large-human-motion-database/
"""
from __future__ import absolute_import
from .classification import *



"""Utility functions"""
# pylint: disable=missing-docstring, arguments-differ, method-hidden
import json
import pickle
import pathlib
import numpy as np


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'to_dict'):
            return obj.to_dict()
        elif isinstance(obj, pathlib.PurePath):
            return str(obj)
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return json.JSONEncoder.default(self, obj)


def load_json(fname, *args, **kwargs):
    with open(fname, 'r') as fd:
        obj = json.load(fd, *args, **kwargs)
    return obj


def save_json(obj, fname, sort_keys=True, indent=4, separators=None,
              encoder=ComplexEncoder, *args, **kwargs):
    with open(fname, 'w') as fd:
        json.dump(obj, fd, indent=indent, sort_keys=sort_keys, cls=encoder,
                  separators=separators, *args, **kwargs)


def load_pickle(fname, **kwargs):
    with open(fname, 'rb') as fd:
        obj = pickle.load(fd, **kwargs)
    return obj


def save_pickle(obj, fname, protocol=None, **kwargs):
    with open(fname, 'wb') as fd:
        pickle.dump(obj, fd, protocol=protocol, **kwargs)


def round_floats_for_json(obj, ndigits=2, key_ndigits=None):
    """
    Tries to round all floats in obj in order to reduce json size.
    ndigits is the default number of digits to round to,
    key_ndigits allows you to override this for specific dictionary keys,
    though there is no concept of nested keys.
    It converts numpy arrays and iterables to lists,
    so it should only be used when serializing to json
    """
    if key_ndigits is None:
        key_ndigits = {}

    if isinstance(obj, np.floating):
        obj = float(obj)
    elif isinstance(obj, np.ndarray):
        obj = obj.tolist()

    if isinstance(obj, float):
        obj = round(obj, ndigits)
    elif isinstance(obj, dict):
        new_obj = {}
        for k, v in obj.items():
            this_ndigits = key_ndigits.get(k, ndigits)
            new_obj[k] = round_floats_for_json(v, this_ndigits, key_ndigits)
        return new_obj
    elif isinstance(obj, str):
        return obj
    else:
        try:
            return [round_floats_for_json(x, ndigits, key_ndigits) for x in obj]
        except TypeError:
            pass

    return obj


def get_encoder_with_float_rounding(ndigits=2, key_ndigits=None):
    class FloatingDigitEncoder(ComplexEncoder):
        def encode(self, obj):
            obj = round_floats_for_json(obj, ndigits, key_ndigits)
            return super().encode(obj)

        def default(self, obj):
            obj = super().default(obj)
            obj = round_floats_for_json(obj, ndigits, key_ndigits)
            return obj

    return FloatingDigitEncoder


Floating2DigitEncoder = get_encoder_with_float_rounding()



from __future__ import print_function
from __future__ import division

import mxnet as mx
import numpy as np

import gluoncv as gcv
from gluoncv.data.batchify import *
from gluoncv.data import DetectionDataLoader, RandomTransformDataLoader


class DummyDetectionDataset(mx.gluon.data.Dataset):
    def __init__(self, size=10):
        self.size = size

    def __len__(self):
        return self.size

    def _fork(self):
        pass

    def __getitem__(self, index):
        num_object = np.random.randint(1, 6)
        return mx.random.normal(shape=(300, 300, 3)), np.random.randn(num_object, 5)

def test_detection_dataloader():
    dataset = DummyDetectionDataset(8)
    for num_workers in [0, 1, 2, 4]:
        for shuffle in (True, False):
            for last_batch in ('keep', 'discard', 'rollover'):
                dataloader = DetectionDataLoader(
                    dataset, batch_size=2, shuffle=shuffle, last_batch=last_batch,
                    num_workers=num_workers)
                for batch in dataloader:
                    mx.nd.waitall()
                    pass

                # new dataloader methods
                batchify_fn = Tuple(Stack(), Pad(pad_val=-1))
                dataloader = mx.gluon.data.DataLoader(
                    dataset, batch_size=2, shuffle=shuffle, last_batch=last_batch,
                    batchify_fn=batchify_fn, num_workers=num_workers)
                for batch in dataloader:
                    mx.nd.waitall()
                    pass

                batchify_fn = Tuple(Append(), Append())
                dataloader = mx.gluon.data.DataLoader(
                    dataset, batch_size=2, shuffle=shuffle, last_batch=last_batch,
                    batchify_fn=batchify_fn, num_workers=num_workers)
                for batch in dataloader:
                    mx.nd.waitall()
                    pass


class DummySequentialDataset(mx.gluon.data.Dataset):
    def __init__(self, size=32, shape=(1,)):
        super(DummySequentialDataset, self).__init__()
        self.size = size
        self.shape = shape

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        return mx.nd.ones(shape=self.shape) * idx

def _fn0(x):
    return x

def _fn1(x):
    return x.tile(reps=(2,))

def _fn2(x):
    return x.tile(reps=(3,))

def test_random_transform_dataloader():
    dataset = DummySequentialDataset(20)
    loader = RandomTransformDataLoader(
        dataset=dataset, shuffle=True, batch_size=4,
        transform_fns=[_fn0, _fn0], last_batch='keep', interval=1, num_workers=2)

    for i in range(4):
        results = []
        for batch in loader:
            results += batch.asnumpy().astype('int').tolist()
            print(batch.asnumpy().astype('int').tolist())

        xx = np.sort(np.array(results).flatten())
        assert (xx == list(range(len(dataset)))).all(), "{}".format(xx)

    # sanity test for rollover
    for last_batch in ['rollover', 'keep', 'discard']:
        loader = RandomTransformDataLoader(
            dataset=dataset, shuffle=True, batch_size=4, transform_fns=[_fn1, _fn2], last_batch=last_batch, interval=2, num_workers=2)
        for i in range(4):
            results = []
            for batch in loader:
                results += batch.asnumpy().astype('int').tolist()

if __name__ == '__main__':
    import nose
    nose.runmodule()



"""Accuracy metric for heatmap prediction."""
# pylint: disable=assignment-from-no-return
import numpy as np
try:
    from mxnet.metric import EvalMetric, check_label_shapes
except ImportError:
    from mxnet.gluon.metric import EvalMetric, check_label_shapes


class HeatmapAccuracy(EvalMetric):
    """Computes heatmap accuracy for keypoint
    Parameters
    ----------
    axis : int, default=1
        The axis that represents classes
    name : str
        Name of this metric instance for display.
    output_names : list of str, or None
        Name of predictions that should be used when updating with update_dict.
        By default include all predictions.
    label_names : list of str, or None
        Name of labels that should be used when updating with update_dict.
        By default include all labels.
    ignore_labels : int or iterable of integers, optional
        If provided as not None, will ignore these labels during update.
    Examples
    --------
    >>> predicts = [mx.nd.array([[0.3, 0.7], [0, 1.], [0.4, 0.6]])]
    >>> labels   = [mx.nd.array([0, 1, 1])]
    >>> acc = mx.metric.Accuracy()
    >>> acc.update(preds = predicts, labels = labels)
    >>> print acc.get()
    ('accuracy', 0.6666666666666666)
    """
    def __init__(self, axis=1, name='heatmap_accuracy', hm_type='gaussian', threshold=0.5,
                 output_names=None, label_names=None, ignore_labels=None):
        super(HeatmapAccuracy, self).__init__(
            name, axis=axis,
            output_names=output_names, label_names=label_names)
        self.axis = axis
        self.ignore_labels = np.array(ignore_labels).flatten()
        self.sum_metric = 0
        self.num_inst = 0
        self.hm_type = hm_type
        self.threshold = threshold

    def _calc_dists(self, preds, target, normalize):
        preds = preds.astype(np.float32)
        target = target.astype(np.float32)
        dists = np.zeros((preds.shape[1], preds.shape[0]))

        for n in range(preds.shape[0]):
            for c in range(preds.shape[1]):
                if target[n, c, 0] > 1 and target[n, c, 1] > 1:
                    normed_preds = preds[n, c, :] / normalize[n]
                    normed_targets = target[n, c, :] / normalize[n]
                    dists[c, n] = np.linalg.norm(normed_preds - normed_targets)
                else:
                    dists[c, n] = -1

        return dists

    def _dist_acc(self, dists):
        dist_cal = np.not_equal(dists, -1)
        num_dist_cal = dist_cal.sum()
        if num_dist_cal > 0:
            return np.less(dists[dist_cal], self.threshold).sum() * 1.0 / num_dist_cal
        else:
            return -1

    def update(self, labels, preds):
        """Updates the internal evaluation result.
        Parameters
        ----------
        labels : list of `NDArray`
            The labels of the data with class indices as values, one per sample.
        preds : list of `NDArray`
            Prediction values for samples. Each prediction value can either be the class index,
            or a vector of likelihoods for all classes.
        """
        from ...data.transforms.pose import get_max_pred
        labels, preds = check_label_shapes(labels, preds, True)

        num_joints = preds[0].shape[1]

        for label, pred in zip(labels, preds):
            norm = 1.0
            h = pred.shape[2]
            w = pred.shape[3]
            if self.hm_type == 'gaussian':
                pred, _ = get_max_pred(pred)
                label, _ = get_max_pred(label)
                norm = np.ones((pred.shape[0], 2)) * np.array([h, w]) / 10

            pred = pred.asnumpy()
            label = label.asnumpy()
            dists = self._calc_dists(pred, label, norm)

            acc = 0
            sum_acc = 0
            cnt = 0

            for i in range(num_joints):
                acc = self._dist_acc(dists[i])
                if acc >= 0:
                    sum_acc += acc
                    cnt += 1

            self.sum_metric += sum_acc
            self.num_inst += cnt



# pylint: disable=missing-function-docstring, line-too-long
"""
Utility functions, misc
"""
import os
import time
import numpy as np


def read_labelmap(labelmap_file):
    """Read label map and class ids."""

    labelmap = []
    class_ids = set()
    name = ""
    class_id = ""
    with open(labelmap_file, 'r') as f:
        for line in f:
            if line.startswith("  name:"):
                name = line.split('"')[1]
            elif line.startswith("  id:") or line.startswith("  label_id:"):
                class_id = int(line.strip().split(" ")[-1])
                labelmap.append({"id": class_id, "name": name})
                class_ids.add(class_id)
    return labelmap, class_ids


def build_log_dir(cfg):
    # create base log directory
    if cfg.CONFIG.LOG.EXP_NAME == 'use_time':
        cfg.CONFIG.LOG.EXP_NAME = time.strftime('%Y-%m-%d-%H-%M-%S', time.localtime(time.time()))
    log_path = os.path.join(cfg.CONFIG.LOG.BASE_PATH, cfg.CONFIG.LOG.EXP_NAME)
    if not os.path.exists(log_path):
        os.makedirs(log_path)

    # dump config file
    with open(os.path.join(log_path, 'config.yaml'), 'w') as f:
        f.write(cfg.dump())

    # create tensorboard saving directory
    tb_logdir = os.path.join(log_path, cfg.CONFIG.LOG.LOG_DIR)
    if not os.path.exists(tb_logdir):
        os.makedirs(tb_logdir)

    # create checkpoint saving directory
    ckpt_logdir = os.path.join(log_path, cfg.CONFIG.LOG.SAVE_DIR)
    if not os.path.exists(ckpt_logdir):
        os.makedirs(ckpt_logdir)

    return tb_logdir


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def calculate_mAP(output, target):
    import torchnet.meter as meter
    mtr = meter.mAPMeter()
    mtr.add(output, target)
    ap = mtr.value()
    return ap


def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


def per_class_error(output, target, num_classes):
    per_class_acc = []
    for i in range(num_classes):
        index = np.where(target == i)
        diff = output[index] - target[index]
        corr = np.where(diff == 0)[0].shape[0] / 1.0
        total = index[0].shape[0] / 1.0
        print(corr, total)
        per_class_acc.append(corr / total)

    return per_class_acc


class ProgressMeter(object):
    """Default PyTorch pogress meter"""
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print('\t'.join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = '{:' + str(num_digits) + 'd}'
        return '[' + fmt + '/' + fmt.format(num_batches) + ']'


def adjust_learning_rate(optimizer, epoch, args):
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    lr = args.lr * (0.1 ** (epoch // 30))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


def get_iou(bb1, bb2):
    """
    Calculate the Intersection over Union (IoU) of two bounding boxes.

    Parameters
    ----------
    bb1 : dict
        Keys: {'x1', 'x2', 'y1', 'y2'}
        The (x1, y1) position is at the top left corner,
        the (x2, y2) position is at the bottom right corner
    bb2 : dict
        Keys: {'x1', 'x2', 'y1', 'y2'}
        The (x, y) position is at the top left corner,
        the (x2, y2) position is at the bottom right corner

    Returns
    -------
    float
        in [0, 1]
    """
    assert bb1['x1'] < bb1['x2']
    assert bb1['y1'] < bb1['y2']
    assert bb2['x1'] < bb2['x2']
    assert bb2['y1'] < bb2['y2']

    # determine the coordinates of the intersection rectangle
    x_left = max(bb1['x1'], bb2['x1'])
    y_top = max(bb1['y1'], bb2['y1'])
    x_right = min(bb1['x2'], bb2['x2'])
    y_bottom = min(bb1['y2'], bb2['y2'])

    if x_right < x_left or y_bottom < y_top:
        return 0.0

    # The intersection of two axis-aligned bounding boxes is always an
    # axis-aligned bounding box
    intersection_area = (x_right - x_left) * (y_bottom - y_top)

    # compute the area of both AABBs
    bb1_area = (bb1['x2'] - bb1['x1']) * (bb1['y2'] - bb1['y1'])
    bb2_area = (bb2['x2'] - bb2['x1']) * (bb2['y2'] - bb2['y1'])

    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    iou = intersection_area / float(bb1_area + bb2_area - intersection_area)
    assert iou >= 0.0
    assert iou <= 1.0
    return iou



"""MobileNet variants with DUC upsampling layers for CenterNet object detection."""
# pylint: disable=unused-argument
from __future__ import absolute_import

import warnings

from mxnet.context import cpu
from mxnet.gluon import nn
from .. model_zoo import get_model
from ...nn.block import DUC

__all__ = ['DUCMobilenet', 'get_duc_mobilenet',
           'mobilenetv3_large_duc', 'mobilenetv3_small_duc']


class DUCMobilenet(nn.HybridBlock):
    """Mobilenet with DUC upsampling block.

    Parameters
    ----------
    base_network : str
        Name of the base feature extraction network.
    up_filters : list of int
        Number of filters for DUC layers.
    up_scales : list of int
        Upsampling scales for DUC layers.
    pretrained_base : bool
        Whether load pretrained base network.
    norm_layer : mxnet.gluon.nn.HybridBlock
        Type of Norm layers, can be BatchNorm, SyncBatchNorm, GroupNorm, etc.
    norm_kwargs : dict
        Additional kwargs for `norm_layer`.

    """
    def __init__(self, base_network='mobilenetv3_small',
                 up_filters=(512, 256, 128), up_scales=(2, 2, 2),
                 pretrained_base=True, norm_layer=nn.BatchNorm, norm_kwargs=None, **kwargs):
        super(DUCMobilenet, self).__init__(**kwargs)
        if norm_layer != nn.BatchNorm:
            raise NotImplementedError('Only standard BatchNorm layer is supported in DUC module')
        assert 'mobilenet' in base_network
        net = get_model(base_network, pretrained=pretrained_base)
        feat = net.features
        idx = [type(l) for l in feat].index(nn.conv_layers.GlobalAvgPool2D)
        with self.name_scope():
            self.feature = feat[:idx]
            self.upsampling = nn.HybridSequential(prefix='upsampling_')
            for f, s in zip(up_filters, up_scales):
                self.upsampling.add(DUC(f, s))

    def hybrid_forward(self, F, x):
        """HybridForward"""
        y = self.feature(x)
        out = self.upsampling(y)
        return out


def get_duc_mobilenet(base_network, pretrained=False, ctx=cpu(), **kwargs):
    """Get mobilenet with duc upsampling layers.

    Parameters
    ----------
    base_network : str
        Name of the base feature extraction network.
    pretrained : bool
        Whether load pretrained base network.
    ctx : mxnet.Context
        mx.cpu() or mx.gpu()
    Returns
    -------
    nn.HybridBlock
        Network instance of mobilenet with duc upsampling layers

    """
    net = DUCMobilenet(base_network=base_network, pretrained_base=pretrained, **kwargs)
    with warnings.catch_warnings(record=True) as _:
        warnings.simplefilter("always")
        net.initialize()
    net.collect_params().reset_ctx(ctx)
    return net

def mobilenetv3_large_duc(**kwargs):
    """Moiblenetv3 large model with duc layers.

    Returns
    -------
    HybridBlock
        A Moiblenetv3 large model with duc layers

    """
    return get_duc_mobilenet('mobilenetv3_large',
                             up_filters=(512, 256, 256), up_scales=(2, 2, 2), **kwargs)

def mobilenetv3_small_duc(**kwargs):
    """Moiblenetv3 small model with duc layers.

    Returns
    -------
    HybridBlock
        A Moiblenetv3 small model with duc layers

    """
    return get_duc_mobilenet('mobilenetv3_small',
                             up_filters=(512, 256, 128), up_scales=(2, 2, 2), **kwargs)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding: utf-8
# pylint: disable=missing-docstring,arguments-differ,unused-argument
"""GoogleNet, implemented in Gluon."""

__all__ = ['GoogLeNet', 'googlenet']

from mxnet.context import cpu
from mxnet.gluon.block import HybridBlock
from mxnet.gluon import nn
from mxnet.gluon.nn import BatchNorm
from mxnet.gluon.contrib.nn import HybridConcurrent

def _make_basic_conv(in_channels, channels, norm_layer=BatchNorm, norm_kwargs=None, **kwargs):
    out = nn.HybridSequential(prefix='')
    out.add(nn.Conv2D(in_channels=in_channels, channels=channels, use_bias=False, **kwargs))
    out.add(norm_layer(in_channels=channels, epsilon=0.001,
                       **({} if norm_kwargs is None else norm_kwargs)))
    out.add(nn.Activation('relu'))
    return out

def _make_branch(use_pool, norm_layer, norm_kwargs, *conv_settings):
    out = nn.HybridSequential(prefix='')
    if use_pool == 'avg':
        out.add(nn.AvgPool2D(pool_size=3, strides=1, padding=1))
    elif use_pool == 'max':
        out.add(nn.MaxPool2D(pool_size=3, strides=1, padding=1))
    setting_names = ['in_channels', 'channels', 'kernel_size', 'strides', 'padding']
    for setting in conv_settings:
        kwargs = {}
        for i, value in enumerate(setting):
            if value is not None:
                if setting_names[i] == 'in_channels':
                    in_channels = value
                elif setting_names[i] == 'channels':
                    channels = value
                else:
                    kwargs[setting_names[i]] = value
        out.add(_make_basic_conv(in_channels, channels, norm_layer, norm_kwargs, **kwargs))
    return out

def _make_Mixed_3a(in_channels, pool_features, prefix, norm_layer, norm_kwargs):
    out = HybridConcurrent(axis=1, prefix=prefix)
    with out.name_scope():
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 64, 1, None, None)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 96, 1, None, None),
                             (96, 128, 3, None, 1)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 16, 1, None, None),
                             (16, 32, 3, None, 1)))
        out.add(_make_branch('max', norm_layer, norm_kwargs,
                             (in_channels, pool_features, 1, None, None)))
    return out

def _make_Mixed_3b(in_channels, pool_features, prefix, norm_layer, norm_kwargs):
    out = HybridConcurrent(axis=1, prefix=prefix)
    with out.name_scope():
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 128, 1, None, None)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 128, 1, None, None),
                             (128, 192, 3, None, 1)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 32, 1, None, None),
                             (32, 96, 3, None, 1)))
        out.add(_make_branch('max', norm_layer, norm_kwargs,
                             (in_channels, pool_features, 1, None, None)))
    return out

def _make_Mixed_4a(in_channels, pool_features, prefix, norm_layer, norm_kwargs):
    out = HybridConcurrent(axis=1, prefix=prefix)
    with out.name_scope():
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 192, 1, None, None)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 96, 1, None, None),
                             (96, 208, 3, None, 1)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 16, 1, None, None),
                             (16, 48, 3, None, 1)))
        out.add(_make_branch('max', norm_layer, norm_kwargs,
                             (in_channels, pool_features, 1, None, None)))
    return out

def _make_Mixed_4b(in_channels, pool_features, prefix, norm_layer, norm_kwargs):
    out = HybridConcurrent(axis=1, prefix=prefix)
    with out.name_scope():
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 160, 1, None, None)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 112, 1, None, None),
                             (112, 224, 3, None, 1)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 24, 1, None, None),
                             (24, 64, 3, None, 1)))
        out.add(_make_branch('max', norm_layer, norm_kwargs,
                             (in_channels, pool_features, 1, None, None)))
    return out

def _make_Mixed_4c(in_channels, pool_features, prefix, norm_layer, norm_kwargs):
    out = HybridConcurrent(axis=1, prefix=prefix)
    with out.name_scope():
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 128, 1, None, None)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 128, 1, None, None),
                             (128, 256, 3, None, 1)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 24, 1, None, None),
                             (24, 64, 3, None, 1)))
        out.add(_make_branch('max', norm_layer, norm_kwargs,
                             (in_channels, pool_features, 1, None, None)))
    return out

def _make_Mixed_4d(in_channels, pool_features, prefix, norm_layer, norm_kwargs):
    out = HybridConcurrent(axis=1, prefix=prefix)
    with out.name_scope():
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 112, 1, None, None)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 144, 1, None, None),
                             (144, 288, 3, None, 1)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 32, 1, None, None),
                             (32, 64, 3, None, 1)))
        out.add(_make_branch('max', norm_layer, norm_kwargs,
                             (in_channels, pool_features, 1, None, None)))
    return out

def _make_Mixed_4e(in_channels, pool_features, prefix, norm_layer, norm_kwargs):
    out = HybridConcurrent(axis=1, prefix=prefix)
    with out.name_scope():
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 256, 1, None, None)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 160, 1, None, None),
                             (160, 320, 3, None, 1)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 32, 1, None, None),
                             (32, 128, 3, None, 1)))
        out.add(_make_branch('max', norm_layer, norm_kwargs,
                             (in_channels, pool_features, 1, None, None)))
    return out

def _make_Mixed_5a(in_channels, pool_features, prefix, norm_layer, norm_kwargs):
    out = HybridConcurrent(axis=1, prefix=prefix)
    with out.name_scope():
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 256, 1, None, None)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 160, 1, None, None),
                             (160, 320, 3, None, 1)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 32, 1, None, None),
                             (32, 128, 3, None, 1)))
        out.add(_make_branch('max', norm_layer, norm_kwargs,
                             (in_channels, pool_features, 1, None, None)))
    return out

def _make_Mixed_5b(in_channels, pool_features, prefix, norm_layer, norm_kwargs):
    out = HybridConcurrent(axis=1, prefix=prefix)
    with out.name_scope():
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 384, 1, None, None)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 192, 1, None, None),
                             (192, 384, 3, None, 1)))
        out.add(_make_branch(None, norm_layer, norm_kwargs,
                             (in_channels, 48, 1, None, None),
                             (48, 128, 3, None, 1)))
        out.add(_make_branch('max', norm_layer, norm_kwargs,
                             (in_channels, pool_features, 1, None, None)))
    return out

def _make_aux(in_channels, classes, norm_layer, norm_kwargs):
    out = nn.HybridSequential(prefix='')
    out.add(nn.AvgPool2D(pool_size=5, strides=3))
    out.add(_make_basic_conv(in_channels=in_channels, channels=128, kernel_size=1,
                             norm_layer=norm_layer, norm_kwargs=norm_kwargs))

    out.add(nn.Flatten())
    out.add(nn.Dense(units=1024, in_units=2048))
    out.add(nn.Activation('relu'))
    out.add(nn.Dropout(0.7))
    out.add(nn.Dense(units=classes, in_units=1024))
    return out

class GoogLeNet(HybridBlock):
    r"""GoogleNet model from
    `"Going Deeper with Convolutions"
    <https://arxiv.org/abs/1409.4842>`_ paper.
    `"Batch Normalization: Accelerating Deep Network Training by Reducing Internal Covariate Shift"
    <https://arxiv.org/abs/1502.03167>`_ paper.

    Parameters
    ----------
    classes : int, default 1000
        Number of classification classes.
    norm_layer : object
        Normalization layer used (default: :class:`mxnet.gluon.nn.BatchNorm`)
        Can be :class:`mxnet.gluon.nn.BatchNorm` or :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.
    norm_kwargs : dict
        Additional `norm_layer` arguments, for example `num_devices=4`
        for :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.
    partial_bn : bool, default False
        Freeze all batch normalization layers during training except the first layer.
    """
    def __init__(self, classes=1000, norm_layer=BatchNorm, dropout_ratio=0.4, aux_logits=False,
                 norm_kwargs=None, partial_bn=False, pretrained_base=True, ctx=None, **kwargs):
        super(GoogLeNet, self).__init__(**kwargs)
        self.dropout_ratio = dropout_ratio
        self.aux_logits = aux_logits

        with self.name_scope():
            self.conv1 = _make_basic_conv(in_channels=3, channels=64, kernel_size=7,
                                          strides=2, padding=3,
                                          norm_layer=norm_layer, norm_kwargs=norm_kwargs)
            self.maxpool1 = nn.MaxPool2D(pool_size=3, strides=2, ceil_mode=True)

            if partial_bn:
                if norm_kwargs is not None:
                    norm_kwargs['use_global_stats'] = True
                else:
                    norm_kwargs = {}
                    norm_kwargs['use_global_stats'] = True

            self.conv2 = _make_basic_conv(in_channels=64, channels=64, kernel_size=1,
                                          norm_layer=norm_layer, norm_kwargs=norm_kwargs)
            self.conv3 = _make_basic_conv(in_channels=64, channels=192,
                                          kernel_size=3, padding=1,
                                          norm_layer=norm_layer, norm_kwargs=norm_kwargs)
            self.maxpool2 = nn.MaxPool2D(pool_size=3, strides=2, ceil_mode=True)

            self.inception3a = _make_Mixed_3a(192, 32, 'Mixed_3a_', norm_layer, norm_kwargs)
            self.inception3b = _make_Mixed_3b(256, 64, 'Mixed_3b_', norm_layer, norm_kwargs)
            self.maxpool3 = nn.MaxPool2D(pool_size=3, strides=2, ceil_mode=True)

            self.inception4a = _make_Mixed_4a(480, 64, 'Mixed_4a_', norm_layer, norm_kwargs)
            self.inception4b = _make_Mixed_4b(512, 64, 'Mixed_4b_', norm_layer, norm_kwargs)
            self.inception4c = _make_Mixed_4c(512, 64, 'Mixed_4c_', norm_layer, norm_kwargs)
            self.inception4d = _make_Mixed_4d(512, 64, 'Mixed_4d_', norm_layer, norm_kwargs)
            self.inception4e = _make_Mixed_4e(528, 128, 'Mixed_4e_', norm_layer, norm_kwargs)
            self.maxpool4 = nn.MaxPool2D(pool_size=2, strides=2)

            self.inception5a = _make_Mixed_5a(832, 128, 'Mixed_5a_', norm_layer, norm_kwargs)
            self.inception5b = _make_Mixed_5b(832, 128, 'Mixed_5b_', norm_layer, norm_kwargs)

            if self.aux_logits:
                self.aux1 = _make_aux(512, classes, norm_layer, norm_kwargs)
                self.aux2 = _make_aux(528, classes, norm_layer, norm_kwargs)

            self.head = nn.HybridSequential(prefix='')
            self.avgpool = nn.AvgPool2D(pool_size=7)
            self.dropout = nn.Dropout(self.dropout_ratio)
            self.output = nn.Dense(units=classes, in_units=1024)
            self.head.add(self.avgpool)
            self.head.add(self.dropout)
            self.head.add(self.output)

    def hybrid_forward(self, F, x):
        x = self.conv1(x)
        x = self.maxpool1(x)

        x = self.conv2(x)
        x = self.conv3(x)
        x = self.maxpool2(x)

        x = self.inception3a(x)
        x = self.inception3b(x)
        x = self.maxpool3(x)

        x = self.inception4a(x)
        if self.aux_logits:
            aux1 = self.aux1(x)
        x = self.inception4b(x)
        x = self.inception4c(x)
        x = self.inception4d(x)
        if self.aux_logits:
            aux2 = self.aux2(x)
        x = self.inception4e(x)
        x = self.maxpool4(x)

        x = self.inception5a(x)
        x = self.inception5b(x)
        x = self.head(x)

        if self.aux_logits:
            return (x, aux2, aux1)
        return x

def googlenet(classes=1000, pretrained=False, pretrained_base=True, ctx=cpu(),
              dropout_ratio=0.4, aux_logits=False,
              root='~/.mxnet/models', partial_bn=False, **kwargs):
    r"""GoogleNet model from
    `"Going Deeper with Convolutions"
    <https://arxiv.org/abs/1409.4842>`_ paper.
    `"Batch Normalization: Accelerating Deep Network Training by Reducing Internal Covariate Shift"
    <https://arxiv.org/abs/1502.03167>`_ paper.

    Parameters
    ----------
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    ctx : Context, default CPU
        The context in which to load the pretrained weights.
    root : str, default $MXNET_HOME/models
        Location for keeping the model parameters.
    partial_bn : bool, default False
        Freeze all batch normalization layers during training except the first layer.
    norm_layer : object
        Normalization layer used (default: :class:`mxnet.gluon.nn.BatchNorm`)
        Can be :class:`mxnet.gluon.nn.BatchNorm` or :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.
    norm_kwargs : dict
        Additional `norm_layer` arguments, for example `num_devices=4`
        for :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.
    """

    net = GoogLeNet(classes=classes, partial_bn=partial_bn, pretrained_base=pretrained_base,
                    dropout_ratio=dropout_ratio, aux_logits=aux_logits, ctx=ctx, **kwargs)
    if pretrained:
        from .model_store import get_model_file
        net.load_parameters(get_model_file('googlenet',
                                           tag=pretrained, root=root), ctx=ctx, cast_dtype=True)
        from ..data import ImageNet1kAttr
        attrib = ImageNet1kAttr()
        net.synset = attrib.synset
        net.classes = attrib.classes
        net.classes_long = attrib.classes_long
    return net



"""MS COCO keypoints dataset."""
from __future__ import absolute_import
from __future__ import division
import os
import copy
import numpy as np
import mxnet as mx
from .utils import try_import_pycocotools
from ..base import VisionDataset
from ...utils.bbox import bbox_xywh_to_xyxy, bbox_clip_xyxy


class COCOKeyPoints(VisionDataset):
    """COCO keypoint detection dataset.

    Parameters
    ----------
    root : str, default '~/mxnet/datasets/coco'
        Path to folder storing the dataset.
    splits : list of str, default ['person_keypoints_val2017']
        Json annotations name.
        Candidates can be: person_keypoints_val2017, person_keypoints_train2017.
    check_centers : bool, default is False
        If true, will force check centers of bbox and keypoints, respectively.
        If centers are far away from each other, remove this label.
    skip_empty : bool, default is False
        Whether skip entire image if no valid label is found. Use `False` if this dataset is
        for validation to avoid COCO metric error.

    """
    CLASSES = ['person']
    KEYPOINTS = {
        0: "nose",
        1: "left_eye",
        2: "right_eye",
        3: "left_ear",
        4: "right_ear",
        5: "left_shoulder",
        6: "right_shoulder",
        7: "left_elbow",
        8: "right_elbow",
        9: "left_wrist",
        10: "right_wrist",
        11: "left_hip",
        12: "right_hip",
        13: "left_knee",
        14: "right_knee",
        15: "left_ankle",
        16: "right_ankle"
    }
    SKELETON = [
        [16, 14], [14, 12], [17, 15], [15, 13], [12, 13], [6, 12], [7, 13], [6, 7], [6, 8],
        [7, 9], [8, 10], [9, 11], [2, 3], [1, 2], [1, 3], [2, 4], [3, 5], [4, 6], [5, 7]]

    def __init__(self, root=os.path.join('~', '.mxnet', 'datasets', 'coco'),
                 splits=('person_keypoints_val2017',), check_centers=False, skip_empty=True):
        super(COCOKeyPoints, self).__init__(root)
        self._root = os.path.expanduser(root)
        if isinstance(splits, mx.base.string_types):
            splits = [splits]
        self._splits = splits
        self._coco = []
        self._check_centers = check_centers
        self._skip_empty = skip_empty
        self.index_map = dict(zip(type(self).CLASSES, range(self.num_class)))
        self.json_id_to_contiguous = None
        self.contiguous_id_to_json = None
        self._items, self._labels = self._load_jsons()

    def __str__(self):
        detail = ','.join([str(s) for s in self._splits])
        return self.__class__.__name__ + '(' + detail + ')'

    @property
    def classes(self):
        """Category names."""
        return type(self).CLASSES

    @property
    def num_joints(self):
        """Dataset defined: number of joints provided."""
        return 17

    @property
    def joint_pairs(self):
        """Joint pairs which defines the pairs of joint to be swapped
        when the image is flipped horizontally."""
        return [[1, 2], [3, 4], [5, 6], [7, 8],
                [9, 10], [11, 12], [13, 14], [15, 16]]

    @property
    def coco(self):
        """Return pycocotools object for evaluation purposes."""
        if not self._coco:
            raise ValueError("No coco objects found, dataset not initialized.")
        if len(self._coco) > 1:
            raise NotImplementedError(
                "Currently we don't support evaluating {} JSON files".format(len(self._coco)))
        return self._coco[0]

    def __len__(self):
        return len(self._items)

    def __getitem__(self, idx):
        img_path = self._items[idx]
        img_id = int(os.path.splitext(os.path.basename(img_path))[0])

        label = copy.deepcopy(self._labels[idx])
        img = mx.image.imread(img_path, 1)
        return img, label, img_id

    def _load_jsons(self):
        """Load all image paths and labels from JSON annotation files into buffer."""
        items = []
        labels = []
        # lazy import pycocotools
        try_import_pycocotools()
        from pycocotools.coco import COCO
        for split in self._splits:
            anno = os.path.join(self._root, 'annotations', split) + '.json'
            _coco = COCO(anno)
            self._coco.append(_coco)
            classes = [c['name'] for c in _coco.loadCats(_coco.getCatIds())]
            if not classes == self.classes:
                raise ValueError("Incompatible category names with COCO: ")
            assert classes == self.classes
            json_id_to_contiguous = {
                v: k for k, v in enumerate(_coco.getCatIds())}
            if self.json_id_to_contiguous is None:
                self.json_id_to_contiguous = json_id_to_contiguous
                self.contiguous_id_to_json = {
                    v: k for k, v in self.json_id_to_contiguous.items()}
            else:
                assert self.json_id_to_contiguous == json_id_to_contiguous
            # iterate through the annotations
            image_ids = sorted(_coco.getImgIds())
            for entry in _coco.loadImgs(image_ids):
                dirname, filename = entry['coco_url'].split('/')[-2:]
                abs_path = os.path.join(self._root, dirname, filename)
                if not os.path.exists(abs_path):
                    raise IOError('Image: {} not exists.'.format(abs_path))
                label = self._check_load_keypoints(_coco, entry)
                if not label:
                    continue

                # num of items are relative to person, not image
                for obj in label:
                    items.append(abs_path)
                    labels.append(obj)
        return items, labels

    def _check_load_keypoints(self, coco, entry):
        """Check and load ground-truth keypoints"""
        ann_ids = coco.getAnnIds(imgIds=entry['id'], iscrowd=False)
        objs = coco.loadAnns(ann_ids)
        # check valid bboxes
        valid_objs = []
        width = entry['width']
        height = entry['height']

        for obj in objs:
            contiguous_cid = self.json_id_to_contiguous[obj['category_id']]
            if contiguous_cid >= self.num_class:
                # not class of interest
                continue
            if max(obj['keypoints']) == 0:
                continue
            # convert from (x, y, w, h) to (xmin, ymin, xmax, ymax) and clip bound
            xmin, ymin, xmax, ymax = bbox_clip_xyxy(bbox_xywh_to_xyxy(obj['bbox']), width, height)
            # require non-zero box area
            if obj['area'] <= 0 or xmax <= xmin or ymax <= ymin:
                continue

            # joints 3d: (num_joints, 3, 2); 3 is for x, y, z; 2 is for position, visibility
            joints_3d = np.zeros((self.num_joints, 3, 2), dtype=np.float32)
            for i in range(self.num_joints):
                joints_3d[i, 0, 0] = obj['keypoints'][i * 3 + 0]
                joints_3d[i, 1, 0] = obj['keypoints'][i * 3 + 1]
                # joints_3d[i, 2, 0] = 0
                visible = min(1, obj['keypoints'][i * 3 + 2])
                joints_3d[i, :2, 1] = visible
                # joints_3d[i, 2, 1] = 0

            if np.sum(joints_3d[:, 0, 1]) < 1:
                # no visible keypoint
                continue

            if self._check_centers:
                bbox_center, bbox_area = self._get_box_center_area((xmin, ymin, xmax, ymax))
                kp_center, num_vis = self._get_keypoints_center_count(joints_3d)
                ks = np.exp(-2 * np.sum(np.square(bbox_center - kp_center)) / bbox_area)
                if (num_vis / 80.0 + 47 / 80.0) > ks:
                    continue

            valid_objs.append({
                'bbox': (xmin, ymin, xmax, ymax),
                'joints_3d': joints_3d
            })

        if not valid_objs:
            if not self._skip_empty:
                # dummy invalid labels if no valid objects are found
                valid_objs.append({
                    'bbox': np.array([-1, -1, 0, 0]),
                    'joints_3d': np.zeros((self.num_joints, 3, 2), dtype=np.float32)
                })
        return valid_objs

    def _get_box_center_area(self, bbox):
        """Get bbox center"""
        c = np.array([(bbox[0] + bbox[2]) / 2.0, (bbox[1] + bbox[3]) / 2.0])
        area = (bbox[3]-bbox[1])*(bbox[2]-bbox[0])
        return c, area

    def _get_keypoints_center_count(self, keypoints):
        """Get geometric center of all keypoints"""
        keypoint_x = np.sum(keypoints[:, 0, 0] * (keypoints[:, 0, 1] > 0))
        keypoint_y = np.sum(keypoints[:, 1, 0] * (keypoints[:, 1, 1] > 0))
        num = float(np.sum(keypoints[:, 0, 1]))
        return np.array([keypoint_x / num, keypoint_y / num]), num



"""GluonCV Utility functions."""
# pylint: disable=wildcard-import
from __future__ import absolute_import

from . import bbox
from . import viz
from . import random
from . import metrics
from . import parallel
from . import filesystem

from .download import download, check_sha1
from .filesystem import makedirs, try_import_dali, try_import_cv2
from .bbox import bbox_iou
from .block import recursive_visit, set_lr_mult, freeze_bn
from .lr_scheduler import LRSequential, LRScheduler
from .plot_history import TrainingHistory
from .export_helper import export_block, export_tvm
from .sync_loader_helper import split_data, split_and_load
from .version import *



import mxnet as mx
from gluoncv.nn import GroupNorm

def test_groupnorm():
    ctx=mx.context.current_context()
    x = mx.nd.random.uniform(1, 2, (4, 16, 8, 8), ctx=ctx)
    gn = GroupNorm(4, 16)
    gn.initialize(ctx=ctx)
    y = gn(x)
    y = y.reshape(0, 4, -1)
    print('y.mean(2) =', y.mean(2))
    mx.test_utils.assert_almost_equal(y.mean(2).asnumpy(),
                                      mx.nd.zeros_like(y.mean(2)).asnumpy(),
                                      rtol=1e-3, atol=1e-3)

if __name__ == '__main__':
    import nose
    nose.runmodule()



# pylint: disable=line-too-long,too-many-lines,missing-docstring
"""Kinetics700 video action recognition dataset.
Code adapted from https://github.com/open-mmlab/mmaction and
https://github.com/bryanyzhu/two-stream-pytorch"""
import os
from ..video_custom import VideoClsCustom

__all__ = ['Kinetics700']

class Kinetics700(VideoClsCustom):
    """Load the Kinetics700 video action recognition dataset.

    Refer to :doc:`../build/examples_datasets/kinetics400` for the description of
    this dataset and how to prepare it.

    Parameters
    ----------
    root : str, required. Default '~/.mxnet/datasets/kinetics400/rawframes_train'.
        Path to the root folder storing the dataset.
    setting : str, required.
        A text file describing the dataset, each line per video sample.
        There are three items in each line: (1) video path; (2) video length and (3) video label.
    train : bool, default True.
        Whether to load the training or validation set.
    test_mode : bool, default False.
        Whether to perform evaluation on the test set.
        Usually there is three-crop or ten-crop evaluation strategy involved.
    name_pattern : str, default None.
        The naming pattern of the decoded video frames.
        For example, img_00012.jpg.
    video_ext : str, default 'mp4'.
        If video_loader is set to True, please specify the video format accordinly.
    is_color : bool, default True.
        Whether the loaded image is color or grayscale.
    modality : str, default 'rgb'.
        Input modalities, we support only rgb video frames for now.
        Will add support for rgb difference image and optical flow image later.
    num_segments : int, default 1.
        Number of segments to evenly divide the video into clips.
        A useful technique to obtain global video-level information.
        Limin Wang, etal, Temporal Segment Networks: Towards Good Practices for Deep Action Recognition, ECCV 2016.
    num_crop : int, default 1.
        Number of crops for each image. default is 1.
        Common choices are three crops and ten crops during evaluation.
    new_length : int, default 1.
        The length of input video clip. Default is a single image, but it can be multiple video frames.
        For example, new_length=16 means we will extract a video clip of consecutive 16 frames.
    new_step : int, default 1.
        Temporal sampling rate. For example, new_step=1 means we will extract a video clip of consecutive frames.
        new_step=2 means we will extract a video clip of every other frame.
    new_width : int, default 340.
        Scale the width of loaded image to 'new_width' for later multiscale cropping and resizing.
    new_height : int, default 256.
        Scale the height of loaded image to 'new_height' for later multiscale cropping and resizing.
    target_width : int, default 224.
        Scale the width of transformed image to the same 'target_width' for batch forwarding.
    target_height : int, default 224.
        Scale the height of transformed image to the same 'target_height' for batch forwarding.
    temporal_jitter : bool, default False.
        Whether to temporally jitter if new_step > 1.
    video_loader : bool, default False.
        Whether to use video loader to load data.
    use_decord : bool, default True.
        Whether to use Decord video loader to load data. Otherwise use mmcv video loader.
    transform : function, default None.
        A function that takes data and label and transforms them.
    slowfast : bool, default False.
        If set to True, use data loader designed for SlowFast network.
        Christoph Feichtenhofer, etal, SlowFast Networks for Video Recognition, ICCV 2019.
    slow_temporal_stride : int, default 16.
        The temporal stride for sparse sampling of video frames in slow branch of a SlowFast network.
    fast_temporal_stride : int, default 2.
        The temporal stride for sparse sampling of video frames in fast branch of a SlowFast network.
    data_aug : str, default 'v1'.
        Different types of data augmentation auto. Supports v1, v2, v3 and v4.
    lazy_init : bool, default False.
        If set to True, build a dataset instance without loading any dataset.
    """
    def __init__(self,
                 root=os.path.expanduser('~/.mxnet/datasets/kinetics700/rawframes_train'),
                 setting=os.path.expanduser('~/.mxnet/datasets/kinetics700/kinetics700_train_list_rawframes.txt'),
                 train=True,
                 test_mode=False,
                 name_pattern='img_%05d.jpg',
                 video_ext='mp4',
                 is_color=True,
                 modality='rgb',
                 num_segments=1,
                 num_crop=1,
                 new_length=1,
                 new_step=1,
                 new_width=340,
                 new_height=256,
                 target_width=224,
                 target_height=224,
                 temporal_jitter=False,
                 video_loader=False,
                 use_decord=False,
                 slowfast=False,
                 slow_temporal_stride=16,
                 fast_temporal_stride=2,
                 data_aug='v1',
                 lazy_init=False,
                 transform=None):

        super(Kinetics700, self).__init__(root, setting, train, test_mode, name_pattern,
                                          video_ext, is_color, modality, num_segments,
                                          num_crop, new_length, new_step, new_width, new_height,
                                          target_width, target_height, temporal_jitter,
                                          video_loader, use_decord, slowfast, slow_temporal_stride,
                                          fast_temporal_stride, data_aug, lazy_init, transform)


class Kinetics700Attr(object):
    def __init__(self):
        self.num_class = 700
        self.classes = ['abseiling', 'acting_in_play', 'adjusting_glasses', 'air_drumming', 'alligator_wrestling', 'answering_questions',
                        'applauding', 'applying_cream', 'archaeological_excavation', 'archery', 'arguing', 'arm_wrestling',
                        'arranging_flowers', 'arresting', 'assembling_bicycle', 'assembling_computer', 'attending_conference',
                        'auctioning', 'baby_waking_up', 'backflip_-human-', 'baking_cookies', 'bandaging', 'barbequing', 'bartending',
                        'base_jumping', 'bathing_dog', 'battle_rope_training', 'beatboxing', 'bee_keeping', 'being_excited',
                        'being_in_zero_gravity', 'belly_dancing', 'bench_pressing', 'bending_back', 'bending_metal', 'biking_through_snow',
                        'blasting_sand', 'blending_fruit', 'blowdrying_hair', 'blowing_bubble_gum', 'blowing_glass', 'blowing_leaves',
                        'blowing_nose', 'blowing_out_candles', 'bobsledding', 'bodysurfing', 'bookbinding', 'bottling',
                        'bouncing_ball_-not_juggling-', 'bouncing_on_bouncy_castle', 'bouncing_on_trampoline', 'bowling', 'braiding_hair',
                        'breading_or_breadcrumbing', 'breakdancing', 'breaking_boards', 'breaking_glass', 'breathing_fire',
                        'brush_painting', 'brushing_floor', 'brushing_hair', 'brushing_teeth', 'building_cabinet', 'building_lego',
                        'building_sandcastle', 'building_shed', 'bulldozing', 'bungee_jumping', 'burping', 'busking', 'calculating',
                        'calligraphy', 'canoeing_or_kayaking', 'capoeira', 'capsizing', 'card_stacking', 'card_throwing', 'carrying_baby',
                        'carrying_weight', 'cartwheeling', 'carving_ice', 'carving_marble', 'carving_pumpkin', 'carving_wood_with_a_knife',
                        'casting_fishing_line', 'catching_fish', 'catching_or_throwing_baseball', 'catching_or_throwing_frisbee',
                        'catching_or_throwing_softball', 'celebrating', 'changing_gear_in_car', 'changing_oil',
                        'changing_wheel_-not_on_bike-', 'chasing', 'checking_tires', 'checking_watch', 'cheerleading', 'chewing_gum',
                        'chiseling_stone', 'chiseling_wood', 'chopping_meat', 'chopping_wood', 'clam_digging', 'clapping',
                        'clay_pottery_making', 'clean_and_jerk', 'cleaning_gutters', 'cleaning_pool', 'cleaning_shoes', 'cleaning_toilet',
                        'cleaning_windows', 'climbing_a_rope', 'climbing_ladder', 'climbing_tree', 'closing_door', 'coloring_in',
                        'combing_hair', 'contact_juggling', 'contorting', 'cooking_chicken', 'cooking_egg', 'cooking_on_campfire',
                        'cooking_sausages_-not_on_barbeque-', 'cooking_scallops', 'cosplaying', 'coughing', 'counting_money',
                        'country_line_dancing', 'cracking_back', 'cracking_knuckles', 'cracking_neck', 'crawling_baby', 'crocheting',
                        'crossing_eyes', 'crossing_river', 'crying', 'cumbia', 'curling_-sport-', 'curling_eyelashes', 'curling_hair',
                        'cutting_apple', 'cutting_cake', 'cutting_nails', 'cutting_orange', 'cutting_pineapple', 'cutting_watermelon',
                        'dancing_ballet', 'dancing_charleston', 'dancing_gangnam_style', 'dancing_macarena', 'deadlifting',
                        'dealing_cards', 'decorating_the_christmas_tree', 'decoupage', 'delivering_mail', 'digging', 'dining',
                        'directing_traffic', 'disc_golfing', 'diving_cliff', 'docking_boat', 'dodgeball', 'doing_aerobics',
                        'doing_jigsaw_puzzle', 'doing_laundry', 'doing_nails', 'doing_sudoku', 'drawing', 'dribbling_basketball',
                        'drinking_shots', 'driving_car', 'driving_tractor', 'drooling', 'drop_kicking', 'drumming_fingers',
                        'dumpster_diving', 'dunking_basketball', 'dyeing_eyebrows', 'dyeing_hair', 'eating_burger', 'eating_cake',
                        'eating_carrots', 'eating_chips', 'eating_doughnuts', 'eating_hotdog', 'eating_ice_cream', 'eating_nachos',
                        'eating_spaghetti', 'eating_watermelon', 'egg_hunting', 'embroidering', 'entering_church', 'exercising_arm',
                        'exercising_with_an_exercise_ball', 'extinguishing_fire', 'faceplanting', 'falling_off_bike', 'falling_off_chair',
                        'feeding_birds', 'feeding_fish', 'feeding_goats', 'fencing_-sport-', 'fidgeting', 'filling_cake',
                        'filling_eyebrows', 'finger_snapping', 'fixing_bicycle', 'fixing_hair', 'flint_knapping', 'flipping_bottle',
                        'flipping_pancake', 'fly_tying', 'flying_kite', 'folding_clothes', 'folding_napkins', 'folding_paper',
                        'front_raises', 'frying_vegetables', 'gargling', 'geocaching', 'getting_a_haircut', 'getting_a_piercing',
                        'getting_a_tattoo', 'giving_or_receiving_award', 'gold_panning', 'golf_chipping', 'golf_driving', 'golf_putting',
                        'gospel_singing_in_church', 'grinding_meat', 'grooming_cat', 'grooming_dog', 'grooming_horse',
                        'gymnastics_tumbling', 'hammer_throw', 'hand_washing_clothes', 'head_stand', 'headbanging', 'headbutting',
                        'helmet_diving', 'herding_cattle', 'high_fiving', 'high_jump', 'high_kick', 'historical_reenactment',
                        'hitting_baseball', 'hockey_stop', 'holding_snake', 'home_roasting_coffee', 'hopscotch', 'hoverboarding',
                        'huddling', 'hugging_-not_baby-', 'hugging_baby', 'hula_hooping', 'hurdling', 'hurling_-sport-', 'ice_climbing',
                        'ice_fishing', 'ice_skating', 'ice_swimming', 'inflating_balloons', 'installing_carpet', 'ironing', 'ironing_hair',
                        'javelin_throw', 'jaywalking', 'jetskiing', 'jogging', 'juggling_balls', 'juggling_fire', 'juggling_soccer_ball',
                        'jumping_bicycle', 'jumping_into_pool', 'jumping_jacks', 'jumping_sofa', 'jumpstyle_dancing', 'karaoke',
                        'kicking_field_goal', 'kicking_soccer_ball', 'kissing', 'kitesurfing', 'knitting', 'krumping', 'land_sailing',
                        'laughing', 'lawn_mower_racing', 'laying_bricks', 'laying_concrete', 'laying_decking', 'laying_stone',
                        'laying_tiles', 'leatherworking', 'letting_go_of_balloon', 'licking', 'lifting_hat', 'lighting_candle',
                        'lighting_fire', 'listening_with_headphones', 'lock_picking', 'long_jump', 'longboarding', 'looking_at_phone',
                        'looking_in_mirror', 'luge', 'lunge', 'making_a_cake', 'making_a_sandwich', 'making_balloon_shapes',
                        'making_bubbles', 'making_cheese', 'making_horseshoes', 'making_jewelry', 'making_latte_art',
                        'making_paper_aeroplanes', 'making_pizza', 'making_slime', 'making_snowman', 'making_sushi', 'making_tea',
                        'making_the_bed', 'marching', 'marriage_proposal', 'massaging_back', 'massaging_feet', 'massaging_legs',
                        'massaging_neck', 'massaging_person-s_head', 'metal_detecting', 'milking_cow', 'milking_goat', 'mixing_colours',
                        'moon_walking', 'mopping_floor', 'mosh_pit_dancing', 'motorcycling', 'mountain_climber_-exercise-', 'moving_baby',
                        'moving_child', 'moving_furniture', 'mowing_lawn', 'mushroom_foraging', 'needle_felting', 'news_anchoring',
                        'opening_bottle_-not_wine-', 'opening_coconuts', 'opening_door', 'opening_present', 'opening_refrigerator',
                        'opening_wine_bottle', 'packing', 'paragliding', 'parasailing', 'parkour', 'passing_American_football_-in_game-',
                        'passing_American_football_-not_in_game-', 'passing_soccer_ball', 'peeling_apples', 'peeling_banana',
                        'peeling_potatoes', 'person_collecting_garbage', 'petting_animal_-not_cat-', 'petting_cat', 'petting_horse',
                        'photobombing', 'photocopying', 'picking_apples', 'picking_blueberries', 'pillow_fight', 'pinching',
                        'pirouetting', 'planing_wood', 'planting_trees', 'plastering', 'playing_accordion', 'playing_american_football',
                        'playing_badminton', 'playing_bagpipes', 'playing_basketball', 'playing_bass_guitar', 'playing_beer_pong',
                        'playing_billiards', 'playing_blackjack', 'playing_cards', 'playing_cello', 'playing_checkers', 'playing_chess',
                        'playing_clarinet', 'playing_controller', 'playing_cricket', 'playing_cymbals', 'playing_darts',
                        'playing_didgeridoo', 'playing_dominoes', 'playing_drums', 'playing_field_hockey', 'playing_flute', 'playing_gong',
                        'playing_guitar', 'playing_hand_clapping_games', 'playing_harmonica', 'playing_harp', 'playing_ice_hockey',
                        'playing_keyboard', 'playing_kickball', 'playing_laser_tag', 'playing_lute', 'playing_mahjong', 'playing_maracas',
                        'playing_marbles', 'playing_monopoly', 'playing_netball', 'playing_nose_flute', 'playing_oboe', 'playing_ocarina',
                        'playing_organ', 'playing_paintball', 'playing_pan_pipes', 'playing_piano', 'playing_piccolo', 'playing_pinball',
                        'playing_ping_pong', 'playing_poker', 'playing_polo', 'playing_recorder', 'playing_road_hockey', 'playing_rounders',
                        'playing_rubiks_cube', 'playing_saxophone', 'playing_scrabble', 'playing_shuffleboard', 'playing_slot_machine',
                        'playing_squash_or_racquetball', 'playing_tennis', 'playing_trombone', 'playing_trumpet', 'playing_ukulele',
                        'playing_violin', 'playing_volleyball', 'playing_with_trains', 'playing_xylophone', 'poaching_eggs',
                        'poking_bellybutton', 'pole_vault', 'polishing_furniture', 'polishing_metal', 'popping_balloons', 'pouring_beer',
                        'pouring_milk', 'pouring_wine', 'preparing_salad', 'presenting_weather_forecast', 'pretending_to_be_a_statue',
                        'pull_ups', 'pulling_espresso_shot', 'pulling_rope_-game-', 'pumping_fist', 'pumping_gas', 'punching_bag',
                        'punching_person_-boxing-', 'push_up', 'pushing_car', 'pushing_cart', 'pushing_wheelbarrow', 'pushing_wheelchair',
                        'putting_in_contact_lenses', 'putting_on_eyeliner', 'putting_on_foundation', 'putting_on_lipstick',
                        'putting_on_mascara', 'putting_on_sari', 'putting_on_shoes', 'putting_wallpaper_on_wall', 'raising_eyebrows',
                        'reading_book', 'reading_newspaper', 'recording_music', 'repairing_puncture', 'riding_a_bike', 'riding_camel',
                        'riding_elephant', 'riding_mechanical_bull', 'riding_mule', 'riding_or_walking_with_horse', 'riding_scooter',
                        'riding_snow_blower', 'riding_unicycle', 'ripping_paper', 'roasting_marshmallows', 'roasting_pig',
                        'robot_dancing', 'rock_climbing', 'rock_scissors_paper', 'roller_skating', 'rolling_eyes', 'rolling_pastry',
                        'rope_pushdown', 'running_on_treadmill', 'sailing', 'salsa_dancing', 'saluting', 'sanding_floor', 'sanding_wood',
                        'sausage_making', 'sawing_wood', 'scrambling_eggs', 'scrapbooking', 'scrubbing_face', 'scuba_diving', 'seasoning_food',
                        'separating_eggs', 'setting_table', 'sewing', 'shaking_hands', 'shaking_head', 'shaping_bread_dough',
                        'sharpening_knives', 'sharpening_pencil', 'shaving_head', 'shaving_legs', 'shearing_sheep', 'shining_flashlight',
                        'shining_shoes', 'shoot_dance', 'shooting_basketball', 'shooting_goal_-soccer-', 'shooting_off_fireworks',
                        'shopping', 'shot_put', 'shouting', 'shoveling_snow', 'shredding_paper', 'shucking_oysters', 'shuffling_cards',
                        'shuffling_feet', 'side_kick', 'sieving', 'sign_language_interpreting', 'silent_disco', 'singing', 'sipping_cup',
                        'situp', 'skateboarding', 'ski_ballet', 'ski_jumping', 'skiing_crosscountry', 'skiing_mono', 'skiing_slalom',
                        'skipping_rope', 'skipping_stone', 'skydiving', 'slacklining', 'slapping', 'sled_dog_racing', 'sleeping',
                        'slicing_onion', 'smashing', 'smelling_feet', 'smoking', 'smoking_hookah', 'smoking_pipe', 'snatch_weight_lifting',
                        'sneezing', 'snorkeling', 'snowboarding', 'snowkiting', 'snowmobiling', 'somersaulting', 'spelunking',
                        'spinning_plates', 'spinning_poi', 'splashing_water', 'spray_painting', 'spraying', 'springboard_diving',
                        'square_dancing', 'squat', 'squeezing_orange', 'stacking_cups', 'stacking_dice', 'standing_on_hands', 'staring',
                        'steer_roping', 'steering_car', 'sticking_tongue_out', 'stomping_grapes', 'stretching_arm', 'stretching_leg',
                        'sucking_lolly', 'surfing_crowd', 'surfing_water', 'surveying', 'sweeping_floor', 'swimming_backstroke',
                        'swimming_breast_stroke', 'swimming_butterfly_stroke', 'swimming_front_crawl', 'swimming_with_dolphins',
                        'swimming_with_sharks', 'swing_dancing', 'swinging_baseball_bat', 'swinging_on_something', 'sword_fighting',
                        'sword_swallowing', 'tackling', 'tagging_graffiti', 'tai_chi', 'taking_photo', 'talking_on_cell_phone',
                        'tango_dancing', 'tap_dancing', 'tapping_guitar', 'tapping_pen', 'tasting_beer', 'tasting_food', 'tasting_wine',
                        'testifying', 'texting', 'threading_needle', 'throwing_axe', 'throwing_ball_-not_baseball_or_American_football-',
                        'throwing_discus', 'throwing_knife', 'throwing_snowballs', 'throwing_tantrum', 'throwing_water_balloon', 'tickling',
                        'tie_dying', 'tightrope_walking', 'tiptoeing', 'tobogganing', 'tossing_coin', 'tossing_salad', 'training_dog',
                        'trapezing', 'treating_wood', 'trimming_or_shaving_beard', 'trimming_shrubs', 'trimming_trees', 'triple_jump',
                        'twiddling_fingers', 'tying_bow_tie', 'tying_knot_-not_on_a_tie-', 'tying_necktie', 'tying_shoe_laces', 'unboxing',
                        'uncorking_champagne', 'unloading_truck', 'using_a_microscope', 'using_a_paint_roller', 'using_a_power_drill',
                        'using_a_sledge_hammer', 'using_a_wrench', 'using_atm', 'using_bagging_machine', 'using_circular_saw',
                        'using_inhaler', 'using_megaphone', 'using_puppets', 'using_remote_controller_-not_gaming-', 'using_segway',
                        'vacuuming_car', 'vacuuming_floor', 'visiting_the_zoo', 'wading_through_mud', 'wading_through_water',
                        'waiting_in_line', 'waking_up', 'walking_on_stilts', 'walking_the_dog', 'walking_through_snow', 'walking_with_crutches',
                        'washing_dishes', 'washing_feet', 'washing_hair', 'washing_hands', 'watching_tv', 'water_skiing', 'water_sliding',
                        'watering_plants', 'waving_hand', 'waxing_armpits', 'waxing_back', 'waxing_chest', 'waxing_eyebrows', 'waxing_legs',
                        'weaving_basket', 'weaving_fabric', 'welding', 'whistling', 'windsurfing', 'winking', 'wood_burning_-art-',
                        'wrapping_present', 'wrestling', 'writing', 'yarn_spinning', 'yawning', 'yoga', 'zumba']
